/*
Copyright (c) 2000,2001. The Regents of the University of California
(Regents). All Rights Reserved. 

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research, and not-for-profit purposes,
without fee and without a signed licensing agreement, is hereby
granted, provided that the above copyright notice, this paragraph and
the following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC
Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620,
(510) 643-7201, for commercial licensing opportunities.
  
IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. 
  
REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF
ANY, PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO
OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.
*/
				 
package orthotweezers;
import java.io.*;
import dk.eos.io.port.IOPort;

/**
 * Interface to the Advantech PCL-711B I/O board with 8 channel 12-bit A/D,
 * 16 digital inputs and 16 digital outputs.
 *
 * @author  unknown
 */
public class AdvantechPCL711B extends Object {

    /** Creates new MultilabPCL711B.  Digital output bytes 0 and 1 are set to 0xff.
     *
     * @param base  base address to which to I/O card is set
     */
    public AdvantechPCL711B(int base) {
        // Create the 16 IOPorts
        for (int i = 0; i < 16; ++i)
            _port[i] = new IOPort (base + i);
        
        // Initialize digital output ports
        out0SetBits (0xff);
        out1SetBits (0xff);
    }

    /** Read digital input byte 0.
     *
     * @return input byte or -1 if there was an I/O error.
     * @see     #in1
     */
    public int in0 () {
        try {
            return (_port[6].in());
        } catch (IOException e) {
            return (-1);
        }
    }

    /** Read digital input byte 1.
     *
     * @return     input byte or -1 if there was an I/O error.
     * @see     #in0
     */
    public int in1 () {
        try {
            return (_port[7].in());
        } catch (IOException e) {
            return (-1);
        }
    }

    /** Set bits on digital output 0 with the bits in bitMask.
     *
     * @param bitMask the bits to turn on
     *
     * @see     #out1SetBits
     */
    public void out0SetBits (int bitMask) {
        _out0State |= bitMask;
        try {
            _port[13].out (_out0State);
        } catch (IOException e) {
            // ignore error
        }
    }

    /** Set bits on digital output 1 with the bits in bitMask.
     *
     * @param bitMask the bits to turn on
     * @see     #out0SetBits
     */
    public void out1SetBits (int bitMask) {
        _out1State |= bitMask;
        try {
            _port[14].out (_out1State);
        } catch (IOException e) {
            // ignore error
        }
    }

    /** Clear bits on digital output 0 with the bits in bitMask.
     *
     * @param bitMask the bits to turn off
     * @see     #out1ClearBits
     */
    public void out0ClearBits (int bitMask) {
        _out0State |= bitMask;
        _out0State -= bitMask;
        try {
            _port[13].out (_out0State);
        } catch (IOException e) {
            // ignore error
        }
    }

    /** Clear bits on digital output 1 with the bits in bitMask.
     *
     * @param bitMask   the bits to turn off
     * @see     #out0ClearBits
     */
    public void out1ClearBits (int bitMask) {
        _out1State |= bitMask;
        _out1State -= bitMask;
        try {
            _port[14].out (_out1State);
        } catch (IOException e) {
            // ignore error
        }
    }
    
    /** Read a 12-bit A/D channel.  The thread is used to sleep for 25 milliseconds
     *  while waiting for the A/D converter.  This is synchronized so that another
     *  thread won't try to start an A/D conversion while this is doing it.
     * 
     * @param channel   the channel to read, from 0 to 7 inclusive.  If the
     *                  channel value is outside this range, this does nothing.
     * @return      the 12-bit A/D value in the range 0 to 0x0fff, or -1 if
     *              the channel number is out of range or there was an
     *              IOException or InterruptedException.
     */
    public synchronized int readAD (int channel) {
        if (channel < 0 || channel > 7)
            return (-1);

        try {
            // mode and interrupt control
            _port[11].out (0x10);
            _port[10].out (channel);
            // trigger the conversion
            _port[12].out (0);
            // Sleep for the initial 25 microseconds while it converts.
            // Sleeping avoids using a tight loop that constantly checks which 
            //   slows down other threads.
            Thread.sleep (0, 25000);

            // Keep reading high byte until success
            int high, low;
            while (true) {
                high = _port[5].in();
                if ((high & 0x10) == 0)
                    // The result is ready
                    break;
            }

            low = _port[4].in() & 0xff;
            return (low + ((high & 0x0f)<<8));
        } catch (IOException e) {
            return (-1);
        } catch (InterruptedException e) {
            return (-1);
        }
    }

    // The 16 ports starting at the base address.
    IOPort _port[] = new IOPort[16];
    
    int _out0State;
    int _out1State;
}