/*
Copyright (c) 2000,2001. The Regents of the University of California
(Regents). All Rights Reserved. 

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research, and not-for-profit purposes,
without fee and without a signed licensing agreement, is hereby
granted, provided that the above copyright notice, this paragraph and
the following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC
Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620,
(510) 643-7201, for commercial licensing opportunities.
  
IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. 
  
REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF
ANY, PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO
OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.
*/
				 
package orthotweezers;

import javax.vecmath.*;

/** GraspBLock is a modal dialog box which performs operation of grasping a block
 * and provides a Cancel button to interrupt it.
 *
 * @author  unknown
 */
public class GraspBlock extends javax.swing.JDialog  implements Runnable {

    /** Creates new modal GraspBlock dialog box.  When the user presses
     *    the Grasp button, this runs an algorithm to grasp a block.  
     *    The algorithm is run in a separate
     *    thread and the user can cancel.
     *
     * @param parent    The parent Frame
     * @param tools   The OrthoTweezersTools used to control the system.  The
     *                  underlying OrthoTweezersControl is obtained from this.
     * @param pallet    The point at the lower left corner on the surface of
     *              the pallet from which the block will be searched for
     * @param edge      The point on the edge of the barrier which will be used
     *              to rotate the block around the grip axis
     */
    public GraspBlock(java.awt.Frame parent, OrthoTweezersTools tools,
                      Point3d pallet, Point3d edge) {
        super (parent, true);
        _tools = tools;
        _control = tools.control();

        _pallet = pallet;
        _edge = edge;
        initComponents ();

        // Start the thread which will call run().
        _cancelThread = new CancelThread (this);
        _cancelThread.start();
    }

    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the FormEditor.
     */
    private void initComponents() {//GEN-BEGIN:initComponents
        cancelButton = new javax.swing.JButton();
        _status = new javax.swing.JLabel();
        getContentPane().setLayout(null);
        setTitle("Grasp Block");
        addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent evt) {
                closeDialog(evt);
            }
        }
        );
        
        cancelButton.setText("Cancel");
        cancelButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                cancelButtonActionPerformed(evt);
            }
        }
        );
        
        getContentPane().add(cancelButton);
        cancelButton.setBounds(30, 10, 90, 20);
        
        
        _status.setText("Running");
        
        getContentPane().add(_status);
        _status.setBounds(10, 40, 230, 20);
        
        setBounds(0, 0, 200, 100);
    }//GEN-END:initComponents

  private void cancelButtonActionPerformed (java.awt.event.ActionEvent evt) {//GEN-FIRST:event_cancelButtonActionPerformed
    // Add your handling code here:
    doClose();
  }//GEN-LAST:event_cancelButtonActionPerformed

private void doClose () {
    // cancel the running thread
    if (_cancelThread != null)
      // cancel the running thread
      _cancelThread.cancel();

    setVisible (false);
    dispose ();
}

/** Closes the dialog */
    private void closeDialog(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_closeDialog
        doClose ();
    }//GEN-LAST:event_closeDialog

    public void run() {
        try {
            _status.setText ("Moving to pallet");
            // Lift up to avoid hitting objects
            _control.moveToZ (_control.stageZ() + .3);
            CancelThread.check();
            _tools.seekYArmDeflection (.1);
            _tools.seekXArmDeflection (0.);
            // The probes should not be touching, so quick calibrate tip offsets
            _control.resetXTipStrain();
            _control.resetYTipStrain();
            
            _control.moveToXY (_pallet.x, _pallet.y);
            _control.moveToZ (_pallet.z + .03);
            _control.moveProbeX (1.);
            CancelThread.check();

            _status.setText ("Locating lower corner of block");
            boolean found = false;
            double y;
            for (y = _pallet.y; y < _pallet.y + 1.; y += .05) {
                _control.moveToXY (_pallet.x, y);
                if (_tools.moveProbeXToContact (false)) {
                    // contacted the block
                    found = true;
                    break;
                }
                // Restore position for next sweep
                _control.moveProbeX (1.);
                CancelThread.check();
            }
            if (!found) {
                // can't contact the block.  Perhaps Z is too high, or we should
                //   move over in X or farther in Y and try again.
                _status.setText ("ERROR: Can't contact block lower corner");
                return;
            }

            Point2d lowerCorner = new Point2d ();
            lowerCorner.x = _control.stageX() + _control.xArmDeflection();
            lowerCorner.y = y;
            // Move the probe a little to sweep again
            _control.moveProbeX (_control.probeXPosition() + .1);

            // back off until we miss the block
            double saveY = y;
            found = false;
            for (y = saveY; y >= saveY - .1; y -= .005) {
                _control.moveToXY (_pallet.x, y);
                if (_tools.moveProbeXToContact (false) == false) {
                    // missed the block
                    found = true;
                    break;
                }
                // Contacted the block again, so revise lowerCorner
                lowerCorner.x = _control.stageX() + _control.xArmDeflection();
                lowerCorner.y = y;
                // Move the probe a little to sweep again
                _control.moveProbeX (_control.probeXPosition() + .1);

                CancelThread.check();
            }
            if (!found) {
                // can't contact the block.  Perhaps Z is too high, or we should
                //   move over in X or farther in Y and try again.
                _status.setText ("ERROR: Can't find block lower corner");
                return;
            }

            // Now do the same operation with probe Y to find the left corner
            _status.setText ("Locating left corner of block");
            _control.moveToXY (_pallet.x, lowerCorner.y - .1);
            _tools.seekXArmDeflection (.1);
            _control.moveProbeY (1.);
            CancelThread.check();
            found = false;
            double x;
            for (x = _pallet.x; x < _pallet.x + 1.; x += .05) {
                _control.moveToXY (x, lowerCorner.y - .1);
                if (_tools.moveProbeYToContact (false)) {
                    // contacted the block
                    found = true;
                    break;
                }
                // Restore position for next sweep
                _control.moveProbeY (1.);
                CancelThread.check();
            }
            if (!found) {
                // can't contact the block.  Perhaps Z is too high, or we should
                //   move over in Y or farther in X and try again.
                _status.setText ("ERROR: Can't contact block left corner");
                return;
            }

            Point2d leftCorner = new Point2d();;
            leftCorner.x = x;
            leftCorner.y = _control.stageY() + _control.yArmDeflection();
            // Move the probe a little to sweep again
            _control.moveProbeY (_control.probeYPosition() + .1);

            // back off until we miss the block
            double saveX = x;
            found = false;
            for (x = saveX; x >= saveX - .1; x -= .005) {
                _control.moveToXY (x, lowerCorner.y - .1);
                if (_tools.moveProbeYToContact (false) == false) {
                    // missed the block
                    found = true;
                    break;
                }
                // Contacted the block again, so revise lowerCorner
                leftCorner.x = x;
                leftCorner.y = _control.stageY() + _control.yArmDeflection();
                // Move the probe a little to sweep again
                _control.moveProbeY (_control.probeYPosition() + .1);

                CancelThread.check();
            }
            if (!found) {
                // can't contact the block.  Perhaps Z is too high, or we should
                //   move over in X or farther in Y and try again.
                _status.setText ("ERROR: Can't find block left corner");
                return;
            }

            _status.setText ("Grasping block");
            // Get a vector lowerRightSide which should point in the direction along the lower
            //   right side up from lowerCorner, which is the same as pointing along the upper
            //   left side from leftCorner.  Do this by setting lowerLeftSide first.
            Vector2d lowerLeftSide = new Vector2d (leftCorner);
            lowerLeftSide.sub (lowerCorner);
            lowerLeftSide.normalize();
            // lowerRightSide is lowerLeftSide rotated by -90 degrees
            Vector2d lowerRightSide = new Vector2d (lowerLeftSide.y, -lowerLeftSide.x);
            // Now get the points of contact
            Point2d probeXContact = new Point2d();
            probeXContact.scaleAdd (.1, lowerRightSide, lowerCorner);
            Point2d probeYContact = new Point2d();
            probeYContact.scaleAdd (.1, lowerRightSide, leftCorner);

            // first, move to clear the block and position probes out of the way
            _control.moveToXY (leftCorner.x - .1, lowerCorner.y - .1);
            _control.moveProbeX (1.);
            _control.moveProbeY (1.);
            CancelThread.check();

            // Position to contact and move the probes in
            // DEBUG: Hack the Y position until we can deal with the "fat finger" problem
            _control.moveToXY (probeYContact.x, probeXContact.y - .02);
            _tools.moveProbeXToContact (false);
            _tools.moveProbeYToContact (false);
            CancelThread.check();

            // Assume part is oriented roughly at 45 degrees
            // Put partial force so that we have a good enough initial grip
            _tools.seekXTipForce (GRIP_FORCE / 4.);
            _tools.seekYTipForce (GRIP_FORCE / 4.);
            // Now, fully grip
            _tools.seekXTipForce (GRIP_FORCE);
            _tools.seekYTipForce (GRIP_FORCE);
            
            // Lift and rotate to 45 degrees
            _control.moveToZ (_control.stageZ() + .3);
            _tools.seekZAxisAngle (Math.PI / 4., GRIP_FORCE);
            
            if (false) {
            _status.setText ("Aligning against edge");
            // edgeContact is where the stage should be for probe X to contact the edge.
            Point2d edgeContact = new Point2d
              (_edge.x - (_control.xArmDeflection() + _control.xTipDeflection()),
               _edge.y);
            // offsetEdgeContact is where the stage should be for the probes to
            //   be offset .05 millimeters from the edge for aligning and rotating.
            // Assume the edge and the block are at PI/4 angle.
            Point2d offsetEdgeContact = new Point2d
              (edgeContact.x - .05 * Math.cos (Math.PI / 4), 
               edgeContact.y - .05 * Math.cos (Math.PI / 4));
            
            // Raise to a Z level below the edge corner so we can push against 
            //   the vertical edge
            _control.moveToZ (_edge.z - .3);
            // Go to a point near offsetEdgeContact but not at it so we get a clean approach
            _control.moveToXY (offsetEdgeContact.x - .5, offsetEdgeContact.y - .5);
            // Now actually touch.  DEBUG: this should be an algorithm that adjusts
            //   angle as the probe tips slide along the block and change deflection
            _control.moveToXY (offsetEdgeContact.x, offsetEdgeContact.y);
            // Now back off from the edge
            _control.moveToXY (offsetEdgeContact.x - .5, offsetEdgeContact.y - .5);

            _status.setText ("Rotating around grip axis");
            // Raise to a Z level above the edge corner
            _control.moveToZ (_edge.z + .3);
            // offsetEdgeContact should allow the offset space for the part to roll
            //    against the edge
            _control.moveToXY (offsetEdgeContact.x, offsetEdgeContact.y);
            // lower against the edge
            }

            _status.setText ("done");
            cancelButton.setText ("Close");
        } catch (CancelException e) {
            // user cancelled.
            _status.setText (e.getMessage());
        }
    }

    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JButton cancelButton;
    private javax.swing.JLabel _status;
    // End of variables declaration//GEN-END:variables

    static final double GRIP_FORCE = 4.;
    private OrthoTweezersTools _tools;
    private OrthoTweezersControl _control;
    private Point3d _pallet;
    private Point3d _edge;
    private CancelThread _cancelThread;
}