/*
Copyright (c) 2000,2001. The Regents of the University of California
(Regents). All Rights Reserved. 

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research, and not-for-profit purposes,
without fee and without a signed licensing agreement, is hereby
granted, provided that the above copyright notice, this paragraph and
the following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC
Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620,
(510) 643-7201, for commercial licensing opportunities.
  
IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. 
  
REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF
ANY, PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO
OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.
*/
				 
package orthotweezers;

/** An OrthoTweezersFrame captures the state of the orthotweezers system at a 
 * certain moment.
 */
public class OrthoTweezersFrame extends Object {

    /** Creates new OrthoTweezersFrame to go to the the current probe and
     *    stage position as given by control.
     *
     * @param control   the OrthoTweezersControl for the system from which the
     *                  current state is read
     */
    public OrthoTweezersFrame(OrthoTweezersControl control) {
        _probeXPosition = control.probeXPosition();
        _probeYPosition = control.probeYPosition();
        _stageX = control.stageX();
        _stageY = control.stageY();
        _stageZ = control.stageZ();
    }
    
    /** Move the orthotweezers system to the state stored in this frame.
     * This first moves the probes.  If one probe has a closing motion while
     * the other has an opening motion, this does the opening motion first.
     * This next moves the stage.  If there is X, Y and Z motion, this does
     * a rising Z motion before moving XY, else the lowering Z motion after XY.
     *
     * @param control   the OrthoTweezersControl for the system
     * @exception CancelException   if user cancelled
     */
    public void moveTo (OrthoTweezersControl control) throws CancelException {
        double deltaProbeX = _probeXPosition - control.probeXPosition();
        double deltaProbeY = _probeYPosition - control.probeYPosition();
        
        // First, move the probes.
        if (deltaProbeX < 0 && deltaProbeY > 0) {
            // Probe X is closing while probe Y is opening, so do probe Y first.
            control.moveProbeY (_probeYPosition);
            control.moveProbeX (_probeXPosition);
        }
        else {
            // Otherwise, probe Y is closing and X is opening, or both are closing
            //   or opening, so do probe X first.
            control.moveProbeX (_probeXPosition);
            control.moveProbeY (_probeYPosition);
        }
        
        // Now, move the stage.
        boolean didZ = false;
        if (_stageZ > control.stageZ()) {
            // raising Z, so move Z first.
            control.moveToZ (_stageZ);
            didZ = true;
        }
        
        control.moveToXY (_stageX, _stageY);
        
        if (!didZ)
            control.moveToZ (_stageZ);
    }

    private double _probeXPosition, _probeYPosition;
    private double _stageX, _stageY, _stageZ;
}