/*
Copyright (c) 2000,2001. The Regents of the University of California
(Regents). All Rights Reserved. 

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research, and not-for-profit purposes,
without fee and without a signed licensing agreement, is hereby
granted, provided that the above copyright notice, this paragraph and
the following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC
Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620,
(510) 643-7201, for commercial licensing opportunities.
  
IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. 
  
REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF
ANY, PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO
OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.
*/
				 
package orthotweezers;

/** OrthoTweezersTools is a utility class for doing more sophisticated operations
 * than are provided by the basic OrthoTweezersControl class.  Logically, high level
 * manipulation routines don't belong in the basic control class which would just
 * clutter it.
 *
 * @see OrthoTweezersControl
 *
 * @author  unknown
 */
public class OrthoTweezersTools extends Object {

    /** Creates new OrthoTweezersTools which uses the given control object.
     *
     * @param control   The OrthoTweezersControl object which the methods will use. 
     */
    public OrthoTweezersTools (OrthoTweezersControl control) {
        _control = control;
    }
    
    /** Returns the OrthoTweezersControl object which was passed to the 
     * constructor.
     */
    public OrthoTweezersControl control() {
        return _control;
    }

    /** Move probe X until its arm deflection is the given value.  This uses
     * feedback from the arm (but not tip) strain gauge.  To move the probe directly
     * without any feedback, use moveProbeX.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param deflection    the deflection to seek to in millimeters
     * @exception CancelException   thrown by CancelThread.check()
     *              or if got an InterruptedException.
     * @see         CancelThread
     * @see OrthoTweezersControl#moveProbeX
     */
    public void seekXArmDeflection (double deflection) throws CancelException {
        try {
            // Do this whole algorithm twice, the first time to get in the
            //   general neighborhood, then wait a little for the probe dynamics
            //   to settle down and then seek again to get closer.
            for (int i = 0; i < 2; ++i) {
                boolean increasing = (_control.xArmDeflection() < deflection);
                double sign = (increasing ? 1. : -1);
                // The first time through, step by more than SEEK_POSITION_INCREMENT to go faster
                for (double position = _control.probeXPosition();
                     increasing ? position < 1. && _control.xArmDeflection() < deflection
                                : position > -1. && _control.xArmDeflection() > deflection;
                     position += sign * (i == 0 ? 3 : 1) * SEEK_POSITION_INCREMENT) {
                    _control.moveProbeX (position);

                    CancelThread.check();
                }

                if (i == 0)
                  Thread.sleep (ARM_SETTLING_MILLIS);
            }
        } catch (InterruptedException e) {
            throw new CancelException ("interrupted exception");
        }
    }

    /** Move probe Y until its arm deflection is the given value.  This uses
     * feedback from the arm (but not tip) strain gauge.  To move the probe directly
     * without any feedback, use moveProbeY.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param deflection    the deflection to seek to in millimeters
     * @exception CancelException   thrown by CancelThread.check()
     *              or if got an InterruptedException.
     * @see         CancelThread
     * @see OrthoTweezersControl#moveProbeY
     */
    public void seekYArmDeflection (double deflection) throws CancelException {
        try {
            // Do this whole algorithm twice, the first time to get in the
            //   general neighborhood, then wait a little for the probe arm dynamics
            //   to settle down and then seek again to get closer.
            for (int i = 0; i < 2; ++i) {
                boolean increasing = (_control.yArmDeflection() < deflection);
                double sign = (increasing ? 1. : -1);
                // The first time through, step by more than SEEK_POSITION_INCREMENT to go faster
                for (double position = _control.probeYPosition();
                     increasing ? position < 1. && _control.yArmDeflection() < deflection
                                : position > -1. && _control.yArmDeflection() > deflection;
                     position += sign * (i == 0 ? 3 : 1) * SEEK_POSITION_INCREMENT) {
                    _control.moveProbeY (position);

                    CancelThread.check();
                }

                if (i == 0)
                  Thread.sleep (ARM_SETTLING_MILLIS);
            }
        } catch (InterruptedException e) {
            throw new CancelException ("interrupted exception");
        }
    }

    /** Move probe X until the tip strain gauge detects a contact force,
     * then back off a little until the force drops near zero.  If increasing
     * is true then the probe position increases, otherwise decreases.
     * To do a similar
     * operation by moving the stage instead of the probe, use moveStageXToContact.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param increasing    true to increase the probe position to contact,
     *                      false to decrease
     * @return  true for success, false if couldn't find contact.
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     * @see     #moveStageXToContact
     */
    public boolean moveProbeXToContact (boolean increasing) throws CancelException {
        double sign = (increasing ? 1. : -1);        
        // Touch with a force 3 times the strain error (converted to force).
        // For increasing probe position, the force is negative
        // Note that it doesn't matter what _strainToForceXTip is, as long as it
        //   is not zero, because seekXTipForce uses the same value, so this
        //   really only reads the strain gauge directly.
        if (!seekXTipForce (-sign * 3. * _control.strainToForceXTip() * _control.strainErrorXTip()))
            return false;
        // Now, release.
        if (!seekXTipForce (0.))
            return false;
        
        return true;
    }

    /** Close probe Y until the tip strain gauge detects a contact force,
     * then back off a little until the force drops near zero.  If increasing
     * is true then the probe position increases, otherwise decreases.
     * To do a similar
     * operation by moving the stage instead of the probe, use moveStageYToContact.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param increasing    true to increase the probe position to contact,
     *                      false to decrease
     * @return  true for success, false if couldn't find contact.
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     * @see         #moveStageYToContact
     */
    public boolean moveProbeYToContact (boolean increasing) throws CancelException {
        double sign = (increasing ? 1. : -1);        
        // Touch with a force 3 times the strain error (converted to force).
        // For increasing probe position, the force is negative
        // Note that it doesn't matter what _strainToForceYTip is, as long as it
        //   is not zero, because seekYTipForce uses the same value, so this
        //   really only reads the strain gauge directly.
        if (!seekYTipForce (-sign * 3. * _control.strainToForceYTip() * _control.strainErrorYTip()))
            return false;
        // Now, release.
        if (!seekYTipForce (0.))
            return false;
        
        return true;
    }

    /** Move the stage in the X direction until the probe X tip strain gauge detects
     * a contact force,
     * then back off a little until the force drops near zero.  If increasing
     * is true then the position increases (as defined by moveToXY), otherwise decreases.
     * If the stage moves
     * the limit distance without contact, then quit.  To do a similar
     * operation by moving the probe instead of the stage, use moveProbeXToContact.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param increasing    true to increase the X position until contact,
     *                      false to decrease
     * @param limit     the maximum distance in millimeters to move while
     *                  checking for contact
     * @return  true for success, false if couldn't find contact before the
     *          limit distance was reached
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     * @see     #moveProbeXToContact
     */
    public boolean moveStageXToContact (boolean increasing, double limit)
    throws CancelException {
        // make sure limit is a positive number
        limit = Math.abs (limit);
        double sign = (increasing ? 1. : -1);
        boolean found = false;
        double startingX = _control.stageX();
        for (double x = startingX;
             increasing ? x < startingX + limit : x > startingX - limit;
             x += sign * MOVE_STAGE_CONTACT_INCREMENT) {
            _control.moveToXY (x, _control.stageY());
            CancelThread.check();

            if (increasing ? _control.xTipStrain() < -3. * _control.strainErrorXTip() :
                             _control.xTipStrain() >  3. * _control.strainErrorXTip()) {
                found = true;
                break;
            }
        }
        if (!found) {
            // couldn't find initial contact
            return false;
        }

        // Now back off on until the strain drops back near zero
        double saveX = _control.stageX();
        for (double x = saveX;
             increasing ? x > saveX - limit : x < saveX + limit;
             x += -sign * MOVE_STAGE_CONTACT_INCREMENT) {
            _control.moveToXY (x, _control.stageY());
            CancelThread.check();

            if (increasing ? _control.xTipStrain() > -_control.strainErrorXTip() :
                             _control.xTipStrain() <  _control.strainErrorXTip()) {
                // Successfully got the contact to release
                return true;
            }
        }

        // Could not get reading to fall back to zero.  (This really shouldn't happen)
        return false;
    }

    /** Move the stage in the Y direction until the probe Y tip strain gauge detects
     * a contact force,
     * then back off a little until the force drops near zero.  If increasing
     * is true then the position increases (as defined by moveToXY), otherwise decreases.
     * If the stage moves
     * the limit distance without contact, then quit.  To do a similar
     * operation by moving the probe instead of the stage, use moveProbeYToContact.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param increasing    true to increase the X position until contact,
     *                      false to decrease
     * @param limit     the maximum distance in millimeters to move while
     *                  checking for contact
     * @return  true for success, false if couldn't find contact before the
     *          limit distance was reached
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     * @see     #moveProbeYToContact
     */
    public boolean moveStageYToContact (boolean increasing, double limit)
          throws CancelException {
        try {
            // make sure limit is a positive number
            limit = Math.abs (limit);
            double sign = (increasing ? 1. : -1);
            boolean found = false;
            double startingY = _control.stageY();
            for (double y = startingY;
                 increasing ? y < startingY + limit : y > startingY - limit;
                 y += sign * MOVE_STAGE_CONTACT_INCREMENT) {
                _control.moveToXY (_control.stageX(), y);
                CancelThread.check();

                // Because the stage shakes the probes, wait for it to settle
                Thread.sleep (ARM_SETTLING_MILLIS);
                CancelThread.check();

                if (increasing ? _control.yTipStrain() < -3. * _control.strainErrorYTip() :
                                 _control.yTipStrain() >  3. * _control.strainErrorYTip()) {
                    found = true;
                    break;
                }
            }
            if (!found) {
                // couldn't find initial contact
                return false;
            }

            // Now back off on until the strain drops back near zero
            double saveY = _control.stageY();
            for (double y = saveY;
                 increasing ? y > saveY - limit : y < saveY + limit;
                 y += -sign * MOVE_STAGE_CONTACT_INCREMENT) {
                _control.moveToXY (_control.stageX(), y);
                CancelThread.check();

                // Probe is in contact so the arms should not vibrate due to
                //   the stage moving so no need to wait for it to settle.
                if (increasing ? _control.yTipStrain() > -_control.strainErrorYTip() :
                                 _control.yTipStrain() <  _control.strainErrorYTip()) {
                    // Successfully got the contact to release
                    return true;
                }
            }

            // Could not get reading to fall back to zero.  (This really shouldn't happen)
            return false;
        } catch (InterruptedException e) {
            throw new CancelException ("interrupted exception");
        }
    }

    /** Move probe X until its tip force is the given value.  This uses
     * feedback from the tip (but not arm) strain gauge.  To move the probe directly
     * without any feedback, use moveProbeX.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param force    the force to seek to in milli Newtons
     * @return  true if achieved the requested force, false if not in which
     *          case the probe is at either position -1 or 1.
     * @exception CancelException   thrown by CancelThread.check()
     *              or if got an InterruptedException.
     * @see         CancelThread
     * @see OrthoTweezersControl#moveProbeX
     */
    public boolean seekXTipForce (double force) throws CancelException {
        try {
            double forceError = _control.strainToForceXTip() * _control.strainErrorXTip();
            
            // Do this whole algorithm twice, the first time to get in the
            //   general neighborhood, then wait a little for the probe dynamics
            //   to settle down and then seek again to get closer.
            for (int i = 0; i < 2; ++i) {
                // arm position must increase to make the tip force go more negative
                boolean increasing = (force < _control.xTipForce());
                double sign = (increasing ? 1. : -1);
                
                boolean found = false;
                // The first time through, step by more than SEEK_POSITION_INCREMENT to go faster
                for (double position = _control.probeXPosition();
                     increasing ? (position < 1.) : (position > -1.);
                     position += sign * (i == 0 ? 3 : 1) * SEEK_POSITION_INCREMENT) {
                    // When testing the tip force, allow success if within force error
                    if (increasing ? _control.xTipForce() < (force + forceError)
                                   : _control.xTipForce() > (force - forceError)) {
                       found = true;
                       break;
                    }
                    _control.moveProbeX (position);

                    CancelThread.check();
                }
                
                if (!found)
                   return false;

                if (i == 0)
                   Thread.sleep (ARM_SETTLING_MILLIS);
            }
        } catch (InterruptedException e) {
            throw new CancelException ("interrupted exception");
        }
            
        return true;
    }

    /** Move probe Y until its tip force is the given value.  This uses
     * feedback from the tip (but not arm) strain gauge.  To move the probe directly
     * without any feedback, use moveProbeY.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param force    the force to seek to in milli Newtons
     * @return  true if achieved the requested force, false if not in which
     *          case the probe is at either position -1 or 1.
     * @exception CancelException   thrown by CancelThread.check()
     *              or if got an InterruptedException.
     * @see         CancelThread
     * @see OrthoTweezersControl#moveProbeY
     */
    public boolean seekYTipForce (double force) throws CancelException {
        try {
            double forceError = _control.strainToForceYTip() * _control.strainErrorYTip();
            
            // Do this whole algorithm twice, the first time to get in the
            //   general neighborhood, then wait a little for the probe dynamics
            //   to settle down and then seek again to get closer.
            for (int i = 0; i < 2; ++i) {
                // arm position must increase to make the tip force go more negative
                boolean increasing = (force < _control.yTipForce());
                double sign = (increasing ? 1. : -1);
                
                boolean found = false;
                // The first time through, step by more than SEEK_POSITION_INCREMENT to go faster
                for (double position = _control.probeYPosition();
                     increasing ? (position < 1.) : (position > -1.);
                     position += sign * (i == 0 ? 3 : 1) * SEEK_POSITION_INCREMENT) {
                    // When testing the tip force, allow success if within force error
                    if (increasing ? _control.yTipForce() < (force + forceError)
                                   : _control.yTipForce() > (force - forceError)) {
                       found = true;
                       break;
                    }
                    _control.moveProbeY (position);

                    CancelThread.check();
                }
                
                if (!found)
                   return false;

                if (i == 0)
                   Thread.sleep (ARM_SETTLING_MILLIS);
            }
        } catch (InterruptedException e) {
            throw new CancelException ("interrupted exception");
        }
            
        return true;
    }

    /** Try to rotate a block about the Z axis to the given angle.
     * This assumes that both probes are in contact with a block which can
     * be rotated freely.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param angle     the angle to rotate to in radians.  This is measured by the relative
     *          positions of the probe tips.  When probe X is at deflection zero
     *          and probe Y has a positive deflection, this angle is zero.
     *          The angle increases to PI/2 where probe Y has zero deflection and probe
     *          X has a positive deflection.  If angle is less than zero or greater than
     *          PI/2, it is clipped to this range.
     * @param gripForce     the gripping force in milliNewtons to maintain while rotating
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     */
    public void seekZAxisAngle (double angle, double gripForce) throws CancelException {
        if (angle < 0.)
            angle = 0.;
        if (angle > Math.PI / 2.)
            angle = Math.PI / 2.;
        
        double positionIncrement = 5. * SEEK_POSITION_INCREMENT;
        
        // DEBUG: should have a failure check to avoid infinite loop.
        for (int i = 0; i < 100; ++i) {
            // Get readings once and reuse it to avoid different readings due to error.
            double xTipStrain = _control.xTipStrain();
            double probeXDeflection = 
              _control.xArmDeflection() + xTipStrain * _control.strainToDeflectionXTip();
            double yTipStrain = _control.yTipStrain();
            double probeYDeflection = 
              _control.yArmDeflection() + yTipStrain * _control.strainToDeflectionYTip();
            
            double currentAngle = Math.atan2 (probeXDeflection, probeYDeflection);
            boolean increasing = (angle > currentAngle);
            
            if (xTipStrain * _control.strainToForceXTip() < gripForce && 
                yTipStrain * _control.strainToForceYTip() < gripForce) {
                // Both probes are under the gripping force, so increase the force (decrease
                //   the deflection) on the probe which would cause rotation in the desired
                //   direction.
                if (increasing) {
                    //seekYArmDeflection (yArmDeflection - deflectionIncrementY);
                    _control.moveProbeY (_control.probeYPosition() - positionIncrement);
                }
                else {
                    //seekXArmDeflection (xArmDeflection - deflectionIncrementX);
                    _control.moveProbeX (_control.probeXPosition() - positionIncrement);
                }
            }
            else {
                // One of the probes is over the gripping force, so decrease the force (increase
                //   the deflection) on the probe which would cause rotation in the desired
                //   direction.
                if (increasing) {
                    //seekXArmDeflection (xArmDeflection + deflectionIncrementX);
                    _control.moveProbeX (_control.probeXPosition() + positionIncrement);
                }
                else {
                    //seekYArmDeflection (yArmDeflection + deflectionIncrementY);
                    _control.moveProbeY (_control.probeYPosition() + positionIncrement);
                }
            }
        }
    }
    
    final static long ARM_SETTLING_MILLIS = 50;
    // For use in moveStageXToContact, etc.  This determines the accuracy
    //   of the contact measurement
    final static double MOVE_STAGE_CONTACT_INCREMENT = .001;
    // For use in seekProbeXDeflection, etc.
    final static double SEEK_POSITION_INCREMENT = .001;

    private OrthoTweezersControl _control;
}