/*
Copyright (c) 2000,2001. The Regents of the University of California
(Regents). All Rights Reserved. 

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research, and not-for-profit purposes,
without fee and without a signed licensing agreement, is hereby
granted, provided that the above copyright notice, this paragraph and
the following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC
Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620,
(510) 643-7201, for commercial licensing opportunities.
  
IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. 
  
REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF
ANY, PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO
OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.
*/
				 
package orthotweezers;

/** Calibrate is a modal dialog box which performs calibration of the tweezers
 * and provides a Cancel button to interrupt it.
 *
 * @author  unknown
 */
public class Calibrate extends ExecuteDialog {

    /** Creates new modal Calibrate dialog box.  When the user presses
     *    the Calibrate button, this runs an algorithm to set the
     *    parameters in tools.control().  The algorithm is run in a separate
     *    thread and the user can cancel.
     *
     * @param parent    The parent Frame
     * @param tools   The OrthoTweezersTools used to control the system.  The
     *                  underlying OrthoTweezersControl is obtained from this.
     */
    public Calibrate(java.awt.Frame parent, OrthoTweezersTools tools) {
        super (parent, "Calibrate", "Calibrate", 
               "Place probe tips together and near barrier and click Calibrate");
        _tools = tools;
        _control = tools.control();
    }

    /** This is called when the new thread is created to execute the procedure.
     */
    public void execute() throws CancelException {
        try {
            double startingZ = _control.stageZ();
            
            // Raise Z so that we won't hit any objects
            setStatus ("Raising Z to avoid objects");
            _control.moveToZ (startingZ + .3);
            CancelThread.check();
            
            // This used instead of findXArmOffset and findYArmOffset
            _control.resetXArmStrain();
            _control.resetYArmStrain();

            // Make sure probes are out of each others' way
            _control.moveProbeX (.75);
            _control.moveProbeY (.75);
            // Get strain error values
            setStatus ("Characterizing strain error");
            // Wait while the probes settle
            Thread.sleep (1000);
            CancelThread.check();
            _control.calibrateStrainErrors();

            // Tips are not touching, so get zero offsets for the tip strain gauges.
            setStatus ("Getting tip zero offsets");
            _control.resetXTipStrain();
            _control.resetYTipStrain();
            CancelThread.check();
  
// Don't use these because we had the user set the tips together            
//            if (!findXArmOffset())
//                return;
//            if (!findYArmOffset())
//                return;            
            
            // Probes should be open.  Restore Z position.
            setStatus ("Restoring Z to original height");
            _control.moveToZ (startingZ);

            if (!findXArmDeflection())
                return;
            if (!findYArmDeflection())
                return;

            // Raise Z again so that we won't hit any objects
            setStatus ("Raising Z to avoid objects");
            _control.moveToZ (startingZ + .3);
            CancelThread.check();
            if (!findXTipDeflectionAndForce())
                return;
            if (!findYTipDeflectionAndForce())
                return;

            // Probes should be open.  Restore Z position.
            setStatus ("Restoring Z to original height");
            _control.moveToZ (startingZ);

            setStatus ("Restoring probe positions");
            _tools.seekXArmDeflection (0.);
            _tools.seekYArmDeflection (0.);

            setStatus ("done");
        } catch (InterruptedException e) {
            // Thread.sleep() threw an exception.
            setStatus (e.getMessage());
        }
    }

    /* Sweep probe X from right to left while slowly moving probe Y
     *   down in order to detect some strain on probe X tip, then
     *   reset the probe X arm position.  
     *   This uses setStatus() to output appropriate messages.
     *   On success, this leaves both probes at position 1.
     *
     * return   true for success, false for error
     */
    private boolean findXArmOffset() throws CancelException {
        setStatus ("Getting probe X arm zero offset");
        _control.moveProbeX (1.);
        _control.moveProbeY (1.);
        boolean found = false;
        for (double yPosition = 1.; yPosition > -1.; yPosition -= .1) {
            _control.moveProbeY (yPosition);

            if (_tools.moveProbeXToContact(false) == true) {
                // We touched probe Y, but maybe only barely so move it a little 
                //   more and try contacting again.
                _control.moveProbeX (1.);
                _control.moveProbeY (yPosition - .1);
                found = _tools.moveProbeXToContact (false);
                break;
            }
            CancelThread.check();

            // Move probe X back to 1 before moving probe Y so probe X won't hit it
            _control.moveProbeX (1.);
            CancelThread.check();
        }
        if (!found) {
            // This really shouldn't happen, but the tips never contacted
            setStatus ("ERROR: Can't get probe X tip contact");
            return false;
        }

        _control.resetXArmStrain();

        // Restore probeX position so it no longer contacts.
        _control.moveProbeX (1.);
        _control.moveProbeY (1.);

        return true;
    }

    /* Sweep probe Y from top to bottom while slowly moving probe X
     *   left in order to detect some strain on probe Y tip, then
     *   reset the probe Y arm position.  
     *   This uses setStatus() to output appropriate messages.
     *
     * return   true for success, false for error
     */
    private boolean findYArmOffset() throws CancelException {
        setStatus ("Getting probe Y arm zero offset");
        // Expect probes to already be at position 1. but set anyway.
        _control.moveProbeX (1.);
        _control.moveProbeY (1.);
        boolean found = false;
        for (double xPosition = 1.; xPosition > -1.; xPosition -= .1) {
            _control.moveProbeX (xPosition);

            if (_tools.moveProbeYToContact(false) == true) {
                // We touched probe X, but maybe only barely so move it a little 
                //   more and try contacting again.
                _control.moveProbeY (1.);
                _control.moveProbeX (xPosition - .1);
                found = _tools.moveProbeYToContact (false);
                break;
            }
            CancelThread.check();

            // Move probe Y back to 1 before moving probe X so probe Y won't hit it
            _control.moveProbeY (1.);
            CancelThread.check();
        }
        if (!found) {
            // This really shouldn't happen, but the tips never contacted
            setStatus ("ERROR: Can't get probe Y tip contact");
            return false;
        }

        _control.resetYArmStrain();

        // Restore probe Y first.
        _control.moveProbeY (1.);
        _control.moveProbeX (1.);

        return true;
    }

    /* Calibrate the X arm deflection by putting the probe at zero position,
     *   hitting an object on the stage, then moving the stage a known distance
     *   away and displacing the probe.  
     *   This assumes that the Z height is already set so that the probes will
     *   contact the calibration block on the stage.  This moves the stage in the
     *   X direction, but restores it to the original position.
     *   This returns the probes to position 1 when finished.
     *   This uses setStatus() to output appropriate messages.
     *
     * return   true for success, false for error
     */
    private boolean findXArmDeflection() throws CancelException {
        setStatus ("Calibrating probe X arm deflection");
        _control.moveProbeX(1.);
        _control.moveProbeY(1.);
        double startingX = _control.stageX();

        // Set an initial approximate deflection parameter just so that
        //   seek to zero will work properly.
        _control.calibrateXArmDeflection (.7);
        _tools.seekXArmDeflection (0.);
        // Move the stage until the calibration block on the stage hits the tip
        //   If we move too far, abort
        if (_tools.moveStageXToContact (false, 1.) == false) {
            // This might happen if there is no calibration block close enough
            setStatus ("ERROR: Can't contact calibration block");
            return false;
        }
            CancelThread.check();

        // Now we know where the contact is.  Move the stage and then move the
        //   probe until it contacts again.
        _control.moveToXY (_control.stageX() + ARM_DEFLECTION_STAGE_OFFSET, _control.stageY());
            CancelThread.check();
        if (_tools.moveProbeXToContact(false) == false) {
            setStatus ("ERROR: Can't get find calibration block");
            return false;
        }
        _control.calibrateXArmDeflection(-ARM_DEFLECTION_STAGE_OFFSET);

        // Restore probe X position and stage X position
        _control.moveProbeX (1.);
        _control.moveToXY (startingX, _control.stageY());

        return true;
    }

    /* Calibrate the Y arm deflection by putting the probe at zero position,
     *   hitting an object on the stage, then moving the stage a known distance
     *   away and displacing the probe.  
     *   This assumes that the Z height is already set so that the probes will
     *   contact the calibration block on the stage.  This moves the stage in the
     *   Y direction, but restores it to the original position.
     *   This returns the probes to position 1 when finished.
     *   This uses setStatus() to output appropriate messages.
     *
     * return   true for success, false for error
     */
    private boolean findYArmDeflection() throws CancelException {
        setStatus ("Calibrating probe Y arm deflection");
        _control.moveProbeX (1.);
        _control.moveProbeY (1.);
        double startingY = _control.stageY();

        // Set an initial approximate deflection parameter just so that
        //   seek to zero will work properly.
        _control.calibrateYArmDeflection (.7);
        _tools.seekYArmDeflection (0.);
        // Move the stage until the calibration block on the stage hits the tip
        //   If we move too far, abort
        if (_tools.moveStageYToContact (false, 1.) == false) {
            // This might happen if there is no calibration block close enough
            setStatus ("ERROR: Can't contact calibration block");
            return false;
        }
            CancelThread.check();

        // Now we know where the contact is.  Move the stage and then move the
        //   probe until it contacts again.
        _control.moveToXY (_control.stageX(), _control.stageY() + ARM_DEFLECTION_STAGE_OFFSET);
        if (_tools.moveProbeYToContact(false) == false) {
            setStatus ("ERROR: Can't get find calibration block");
            return false;
        }
        _control.calibrateYArmDeflection(-ARM_DEFLECTION_STAGE_OFFSET);

        // Return to original position
        _control.moveProbeY (1.);
        _control.moveToXY (_control.stageX(), startingY);

        return true;
    }

    /* Move probe X to contact probe Y, then push farther and calibrate the
     *   tip deflection as the negative of the arm deflection.  Also
     *   calibrate the tip force.  
     *   This uses setStatus() to output appropriate messages.
     *   On success, this leaves both probes at position 1.
     *
     * return   true for success, false for error
     */
    private boolean findXTipDeflectionAndForce() throws CancelException {
        setStatus ("Calibrating probe X tip deflection");

        // Move probe Y so probe X can touch it
        _tools.seekYArmDeflection (-0.1);
        CancelThread.check();
        // Bend the tip.
        _tools.seekXArmDeflection (-TIP_DEFLECTION_PROBE_OFFSET);
        _control.calibrateXTipDeflection (TIP_DEFLECTION_PROBE_OFFSET);
        // For now, calibrate the force to be the same as the strain.  
        _control.calibrateXTipForce (_control.xTipStrain());
        CancelThread.check();

        // Restore.
        _control.moveProbeX (1.);
        _control.moveProbeY (1.);

        return true;
    }

    /* Move probe Y to contact probe X, then push farther and calibrate the
     *   tip deflection as the negative of the arm deflection.  Also
     *   calibrate the tip force.
     *   This uses setStatus() to output appropriate messages.
     *   On success, this leaves both probes at position 1.
     *
     * return   true for success, false for error
     */
    private boolean findYTipDeflectionAndForce() throws CancelException {
        setStatus ("Calibrating probe Y tip deflection");

        // Move probe Y so probe X can touch it
        _tools.seekXArmDeflection (-0.1);
        CancelThread.check();
        // Bend the tip.
        _tools.seekYArmDeflection (-TIP_DEFLECTION_PROBE_OFFSET);
        _control.calibrateYTipDeflection (TIP_DEFLECTION_PROBE_OFFSET);
        // For now, calibrate the force to be the same as the strain.  
        _control.calibrateYTipForce (_control.yTipStrain());
        CancelThread.check();

        // Restore.  Move probe Y first.
        _control.moveProbeY (1.);
        _control.moveProbeX (1.);

        return true;
    }
    private OrthoTweezersControl _control;
    private OrthoTweezersTools _tools;
    static final double ARM_DEFLECTION_STAGE_OFFSET = .5;
    static final double TIP_DEFLECTION_PROBE_OFFSET = .05;
}