/*
Copyright (c) 2000,2001. The Regents of the University of California
(Regents). All Rights Reserved.
 
Permission to use, copy, modify, and distribute this software and its
documentation for educational, research, and not-for-profit purposes,
without fee and without a signed licensing agreement, is hereby
granted, provided that the above copyright notice, this paragraph and
the following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC
Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620,
(510) 643-7201, for commercial licensing opportunities.
 
IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
 
REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF
ANY, PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO
OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.
 */

package orthotweezers;

import java.io.*;
import org.python.util.*; 
import org.python.core.*; 

/** A CancelInteractiveConsole extends the jython InteractiveConsole to override
 * methods allowing it to handle CancelException while running the interact()
 * method.  It runs the read requests in a separate thread so that the calling
 * thread can cancel.  While this is an added complexity, it is necessary because
 * it is not normally possible to cancel a request to read the input once it is
 * started.
 *
 * @author  unknown
 */
public class CancelInteractiveConsole extends InteractiveConsole {

    /** Creates new CancelInteractiveConsole.  This just calls super(). */
    public CancelInteractiveConsole() {
        super();
        
        // Create a new BufferedReaderThread for reading System.in which starts
        //   immediately and is used whenever raw_input wants to read a line.
        _readerThread = new BufferedReaderThread
          (new BufferedReader (new InputStreamReader (System.in)));
    }

    /** This can be called before interact() to make sure any previous
     * lines in the input buffer are cleared.  This is useful because if
     * the user clicked cancel the previous time while a line was being typed,
     * we don't want it included in the next interactive session.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     */
    public void flushReader() throws CancelException {
        try {
            if (!_readerThread.isReady()) {
                // The user clicked cancel previously while typing a line,
                //   but did not hit enter, so wait for the user to clear it.
                System.out.println("\nPlease hit Enter in the console to clear the input from the previous Cancel.");
                while (!_readerThread.isReady()) {
                    CancelThread.check();
                    Thread.sleep (100L);
                }
            }
            
            // Now, we initiate the readLine and check if there is a response within
            //   100 milliseconds. If so, it means the result came from buffered text which
            //   the user has been typing before now, so we ignore it.
            boolean isFlushMessagePrinted = false;
            while (true) {
                _readerThread.readLine();
                Thread.sleep (2 * BufferedReaderThread.CHECK_REQUEST_MILLIS);
                if (!_readerThread.isReady())
                    // We don't have a result, so assume the user is really typing.
                    // This will be caught by raw_input().
                    break;
                
                // We already have a result, which must have been buffered, so read again.
                if (!isFlushMessagePrinted) {
                    System.out.println("Flushing previously buffered input.");
                    isFlushMessagePrinted = true;
                }
            }
        } catch (InterruptedException e) {
            // Treat this like a CancelException
            throw new CancelException ("operation interrupted");
        }
    }
    
    /** Write a prompt and read a line.  When the
     * user enters "exit" or the input stream ends, EOFError is raised. If cancelThread.check()
     * throws an exception, this raises it as a PyException.
     * This overrides raw_input in InteractiveConsole.
     *
     * @param prompt    the prompt to print for reading the line
     * @return the line read from the input. The returned line does not include 
     *          the trailing newline.  
     * @exception PyException  For the end of input, the type of exception is
     *          Py.EOFError, which will be caught by the interact() method
     *          causing it to exit.  For a cancel returned from CancelThread.check(),
     *          this converts the exception to a PyException with Py.JavaError,
     *          which will be re-thrown by interact().
     */
    public String raw_input(PyObject prompt) {
        String line;
        try {
            System.out.print(prompt.toString());
            
            // Since readLine() cannot be interrupted when the user clicks
            //   cancel, we have to go through the trouble of using the
            //   BufferedReaderThread which calls readLine() on the input in
            //   a separate thread. 
            // Note that readLine() may already have been called by flushReader(),
            //   which is okay.
            _readerThread.readLine();
            while (!_readerThread.isReady()) {
                CancelThread.check();
                Thread.sleep (2 * BufferedReaderThread.CHECK_REQUEST_MILLIS);
            }
            
            line = _readerThread.getReadLineResult();
        } catch (CancelException e) {
            // Convert CancelException to a Python error
            throw Py.JavaError (e);
        } catch (InterruptedException e) {
            // Treat this like CancelException
            throw Py.JavaError (new CancelException ("operation interrupted"));
        }
        
        if (line == null || line.equalsIgnoreCase("exit"))
            throw new PyException (Py.EOFError, Py.EOFError);
        
        return line;
    }

    /** Execute a code object. All exceptions are caught and simply displayed with
     * showexception() except Py.SystemExit and 
     * a PyException version of CancelException, which are re-raised.
     * This overrides InteractiveCompiler.runcode() which is called by runsource()
     * which is called by push() which is called by interact(). This does the
     * exact same thing except also re-raises the PyException version of
     * a CancelException.
     *
     * @param code  the code object created by runsource.
     * @exception PyException  The type of the exception is Py.SystemExit
     *      or the PyException version of CancelException.
     */
    public void runcode(PyObject code) {
        try {
            exec(code);
        } catch (PyException pythonException) {
            if (Py.matchException(pythonException, Py.SystemExit)) 
                throw pythonException;
            if (pythonException.value.__tojava__(Object.class) instanceof CancelException)
                // This is a PyException version of CancelException so re-raise it.
                throw pythonException;

            // Otherwise just show the exception and return cleanly.
            showexception(pythonException);
        }
    }

    private BufferedReaderThread _readerThread;
}

/** This class creates a separate thread which waits for requests to do operations
 * on a supplied BufferedReader.  The reason you may want to run I/O operations
 * on the BufferedReader in a separate thread is that they cannot be suspended or treated with
 * CancelThread.check().
 */
class BufferedReaderThread extends Thread {
    /** Create a new BufferedReaderThread which will provide access to the given reader.
     * The thread which monitors read requests is automatically started.
     */
    BufferedReaderThread (BufferedReader reader) {
        super();
        _reader = reader;
        
        // Start the thread now.  Assume it will terminate when the
        //   Java virtual machine exits.
        start();
    }
    
    /** This this is called by the thread which created this BufferedReaderThread
     * to make this call readLine() on the given reader. The calling thread
     * should repeatedly call isReady() to wait for the result to be available.
     * When the result is ready, use getReadLineResult().  If isReady() is already
     * false (because a previous read request was made but not finished), this
     * does nothing.
     *
     * @see #isReady
     */
    void readLine() {
        if (_request != 0)
            // Already doing a request
            return;
        
        _request = READ_LINE_REQUEST;
    }
    
    /** Return true if ready for a new read request and the result of  
     * a previous requested operation is available. 
     */
    boolean isReady() {
        return (_request == 0);
    }
    
    /** Get the result of the readLine operation which was initiated by calling
     * readLine().  You should call isReady() first to make sure this result is 
     * avaialable.
     *
     * @see #isReady
     */
    String getReadLineResult() {
        return _readLineResult;
    }
    
    /** Watch _request for request operations. When a request appears, do it
     * and set _avaliable true.
     */
    public void run() {
        while (true) {
            if (_request == READ_LINE_REQUEST) {
                try {
                  _readLineResult = _reader.readLine();
                } catch (IOException e) {
                    System.out.println("ERROR in BufferedReaderThread " + e);
                    continue;
                }
                
                _request = 0;
            }

            // Sleep so we don't tie up resources
            try {
                Thread.sleep(CHECK_REQUEST_MILLIS);
            } catch (InterruptedException e) {
                // Ignore
            }
        }
    }
    
    private BufferedReader _reader;
    private int _request;
    private String _readLineResult = "";
    
    private static int READ_LINE_REQUEST = 1;
    
    /** This is the period in milliseconds with which the thread checks for
     * new read requests.
     */ 
    static public long CHECK_REQUEST_MILLIS = 50L;
}