/*
Copyright (c) 2000,2001. The Regents of the University of California
(Regents). All Rights Reserved. 

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research, and not-for-profit purposes,
without fee and without a signed licensing agreement, is hereby
granted, provided that the above copyright notice, this paragraph and
the following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC
Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620,
(510) 643-7201, for commercial licensing opportunities.
  
IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. 
  
REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF
ANY, PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO
OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.
*/
				 
package orthotweezers;

/** A CancelThread extends Thread and allows a mechanism to cleanly pause or cancel the thread.
 * When the thread is created its state is ENABLED (not to be confused with
 * the interrupted or alive state).  If the pause() or cancel() method is called, the 
 * state is set to PAUSED or CANCELLED, respectively.  A thread may call CancelThread.check() 
 * and if pause() has been called (by another thread), check() will loop until the other 
 * thread calls unpause() or cancel().  If cancelled, check() throws a CancelException.
 *
 * @author  unknown
 */

public class CancelThread extends Thread {
    /** Creates a new CancelThread as in Thread(Runnable target)
     */
    public CancelThread (Runnable runnable) {
        super (runnable);
    }
    
    /** This sets the state to CANCELLED so that the next call to check()
     * from the running thread will throw a CancelException, or if check() is
     * looping in the PAUSED state then it will also throw a CancelException.  Once the 
     * CancelException is thrown,
     * the run() method should exit causing the thread to stop running.
     * This also uses isAlive() to wait until the cancelled thread is dead, so  
     * you know that when cancel() returns the cancelled thread is really dead.  Note that
     * if for some reason, the cancelled thread does not die, this method
     * may not return.  
     *
     * @see #check
     */
    public void cancel() {
        _state = CANCELLED;
        // Note that the thread which calls cancel() is not the thread
        //   defined by this.
        while (this.isAlive());
    }
    
    /** This sets the state to PAUSE_REQUEST so that the next call to check()
     * from the running thread will convert the state to PAUSED and loop until
     * unpause() or cancel() is called.  However, if the state is already PAUSED
     * or CANCELLED, this simply returns without doing anything.
     * This waits until the check() method responds and converts the PAUSE_REQUEST
     * to PAUSED, so you know that when this returns, the paused thread is really 
     * pausing.  Note that if for some reason, the other thread does not call
     * check() in order to become paused, this method may not return.
     *
     * @see #check
     * @see #unpause
     * @see #cancel
     */
    public void pause() {
        if (_state == PAUSED || _state == CANCELLED)
            // Can't pause if already paused or cancelled.
            return;
        
        _state = PAUSE_REQUEST;
        
        while (_state == PAUSE_REQUEST);
    }
    
    /** If the state is PAUSED, this sets the state back to ENABLED so that
     * the check() method in which the other thread is looping can exit and
     * allow the thread to resume.  If the state is alreay ENABLED, this does
     * nothing.  Also, if the state is already CANCELLED, this also does nothing
     * because you can't unpause a cancelled thread.
     *
     * @see #check
     * @see #pause
     */
    public void unpause() {
        if (_state != PAUSED)
            // Can't unpause if already enabled or cancelled.
            return;
        
        _state = ENABLED;
    }
    
    /** If this thread state is still ENABLED, simply return.  
     * If the state is CANCELLED (because another thread called cancel()) then this throws
     * a CancelException.  If the state is PAUSED or PAUSE_REQUEST (because another
     * thread called pause()) then this will loop until the state becomes ENABLED
     * (because another thread called unpause()) or CANCELLED.
     * <p>This is a static method and queries Thread.currentThread()
     * to see if the current thread is an instanceof CancelThread.  If not, it
     * simply returns, but if it is, then it looks at the state as described.
     *
     * @see #cancel
     * @see #pause
     * @see #unpause
     * @exception CancelException   if this object is no longer enabled
     */
    public static void check() throws CancelException {
        if (!(Thread.currentThread() instanceof CancelThread))
            return;
            
        CancelThread currentThread = (CancelThread)Thread.currentThread();
        
        if (currentThread._state == PAUSE_REQUEST)
            // Convert a PAUSED_REQUEST into PAUSED to signal that we got it.
            currentThread._state = PAUSED;
        
        // This gets the state and loops as long as it is PAUSED.  Note that
        // If the state is already ENABLED or CANCELLED, this loop will break
        // immediately.
        while (true) {
            if (currentThread._state != PAUSED)
                break;
            
            try {
                // Wait a little while pausing so as not to consume the CPU
                Thread.sleep (1);
            } catch (InterruptedException e) {
                // Treat this as a cancel
                throw new CancelException ("operation cancelled");
            }
        }

        if (currentThread._state == CANCELLED)
            throw new CancelException ("operation cancelled");
        
        // Assume state is not CANCELLED or PAUSED, so it is ENABLED, so return.
    }
    
    /** Return true if the state of this thread is PAUSED.
     * Note: to find out if the thread is cancelled, use the isAlive() method
     * inherited from the Thread class.
     */
    public boolean isPaused() {
        return (_state == PAUSED);
    }
    
    public static int ENABLED = 0;
    public static int PAUSE_REQUEST = 1;
    public static int PAUSED = 2;
    public static int CANCELLED = 3;
    private int _state = ENABLED;
}
