/*
Copyright (c) 2000,2001. The Regents of the University of California
(Regents). All Rights Reserved. 

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research, and not-for-profit purposes,
without fee and without a signed licensing agreement, is hereby
granted, provided that the above copyright notice, this paragraph and
the following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC
Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620,
(510) 643-7201, for commercial licensing opportunities.
  
IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. 
  
REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF
ANY, PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO
OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.
*/
				 
package orthotweezers;

import java.io.*;

/**
 * OrthoTweezersControl interfaces with all the hardware of the orthotweezer
 * system, providing control of the stage, tweezers, strain gauges, etc.
 * This class contains only the basic methods necessary to interface to the
 * system.  For more sophisticated manipulation routines, see OrthoTweezersTools
 * which uses this class.
 *
 * @author  unknown
 * @see OrthoTweezersTools
 */
public class OrthoTweezersControl {

    /** Creates new OrthoTweezersControl.
     * Before exiting, the application should call reset() to return the orthotweezers
     * system to its default state.
     *
     *@see #reset
     */
    public OrthoTweezersControl() {
        _motorAndGauges = new AdvantechPCL711B (0x220);
        // All digital outputs are 1 by default, so set relay to zero.  This
        //   happens quickly enough that the relay should not have time to engage.
        relay1Off();
        
        // Initialize D/A outputs to 0x07ff which is zero volts.
        _analogOutputs = new AdvantechPCL726 (0x2c0, 0x07ff);
    }

    /** Initialize control parameters, such as stageX, from the preferences.
     *
     * @param preferences   the Preferences object which should already have
     *                      values loaded from the preferences file
     */
    public void setFromPreferences (Preferences preferences) {
        // start with initial values.
        double x = stageX(), y = stageY(), z = stageZ();

        x = preferences.maybeGetDouble (Preferences.STAGE_X, x);
        y = preferences.maybeGetDouble (Preferences.STAGE_Y, y);
        z = preferences.maybeGetDouble (Preferences.STAGE_Z, z);
        setStageXYZ (x, y, z);

        _strainErrorXArm = preferences.maybeGetDouble
        (Preferences.STRAIN_ERROR_X_ARM, _strainErrorXArm);
        _strainErrorYArm = preferences.maybeGetDouble
        (Preferences.STRAIN_ERROR_Y_ARM, _strainErrorYArm);
        _strainErrorXTip = preferences.maybeGetDouble
        (Preferences.STRAIN_ERROR_X_TIP, _strainErrorXTip);
        _strainErrorYTip = preferences.maybeGetDouble
        (Preferences.STRAIN_ERROR_Y_TIP, _strainErrorYTip);

        _strainOffsetXArm = preferences.maybeGetDouble
        (Preferences.STRAIN_OFFSET_X_ARM, _strainOffsetXArm);
        _strainOffsetYArm = preferences.maybeGetDouble
        (Preferences.STRAIN_OFFSET_Y_ARM, _strainOffsetYArm);
        _strainOffsetXTip = preferences.maybeGetDouble
        (Preferences.STRAIN_OFFSET_X_TIP, _strainOffsetXTip);
        _strainOffsetYTip = preferences.maybeGetDouble
        (Preferences.STRAIN_OFFSET_Y_TIP, _strainOffsetYTip);

        _strainToDeflectionXArm = preferences.maybeGetDouble
        (Preferences.STRAIN_TO_DEFLECTION_X_ARM, _strainToDeflectionXArm);
        _strainToDeflectionYArm = preferences.maybeGetDouble
        (Preferences.STRAIN_TO_DEFLECTION_Y_ARM, _strainToDeflectionYArm);
        _strainToDeflectionXTip = preferences.maybeGetDouble
        (Preferences.STRAIN_TO_DEFLECTION_X_TIP, _strainToDeflectionXTip);
        _strainToDeflectionYTip = preferences.maybeGetDouble
        (Preferences.STRAIN_TO_DEFLECTION_Y_TIP, _strainToDeflectionYTip);
        _strainToForceXTip = preferences.maybeGetDouble
        (Preferences.STRAIN_TO_FORCE_X_TIP, _strainToForceXTip);
        _strainToForceYTip = preferences.maybeGetDouble
        (Preferences.STRAIN_TO_FORCE_Y_TIP, _strainToForceYTip);
    }

    /** Set the relevant fields in the Preferences object, such as STAGE_X,
     * with the values in this object.  Typically an application calls
     * getPreferences just before saving the preferences to a file.
     *
     * @param preferences   the Preferences object into which the control
     *                      parameters will be saved
     */
    public void getPreferences (Preferences preferences) {
        preferences.putDouble (Preferences.STAGE_X, stageX());
        preferences.putDouble (Preferences.STAGE_Y, stageY());
        preferences.putDouble (Preferences.STAGE_Z, stageZ());

        preferences.putDouble (Preferences.STRAIN_ERROR_X_ARM, _strainErrorXArm);
        preferences.putDouble (Preferences.STRAIN_ERROR_Y_ARM, _strainErrorYArm);
        preferences.putDouble (Preferences.STRAIN_ERROR_X_TIP, _strainErrorXTip);
        preferences.putDouble (Preferences.STRAIN_ERROR_Y_TIP, _strainErrorYTip);

        preferences.putDouble (Preferences.STRAIN_OFFSET_X_ARM, _strainOffsetXArm);
        preferences.putDouble (Preferences.STRAIN_OFFSET_Y_ARM, _strainOffsetYArm);
        preferences.putDouble (Preferences.STRAIN_OFFSET_X_TIP, _strainOffsetXTip);
        preferences.putDouble (Preferences.STRAIN_OFFSET_Y_TIP, _strainOffsetYTip);

        preferences.putDouble (Preferences.STRAIN_TO_DEFLECTION_X_ARM, _strainToDeflectionXArm);
        preferences.putDouble (Preferences.STRAIN_TO_DEFLECTION_Y_ARM, _strainToDeflectionYArm);
        preferences.putDouble (Preferences.STRAIN_TO_DEFLECTION_X_TIP, _strainToDeflectionXTip);
        preferences.putDouble (Preferences.STRAIN_TO_DEFLECTION_Y_TIP, _strainToDeflectionYTip);
        preferences.putDouble (Preferences.STRAIN_TO_FORCE_X_TIP, _strainToForceXTip);
        preferences.putDouble (Preferences.STRAIN_TO_FORCE_Y_TIP, _strainToForceYTip);
    }

    /** Reset the orthotweezers system to its default state.
     * This moves probe X and probe Y to the zero position.
     * This should be called before exiting the application.
     *
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     */
    public void reset() throws CancelException {
        moveProbeX (0);
        moveProbeY (0);
        relay1Off();
    }

    /** Move X one step, updating _xSteps.
     *
     * @param direction     true for increasing X, false for decreasing
     */
    public void xStep (boolean direction)
    {
        if (direction) {
            _motorAndGauges.out0SetBits (X_DIRECTION);
            _xSteps++;
        }
        else {
            _motorAndGauges.out0ClearBits (X_DIRECTION);
            _xSteps--;
        }

        try {
            // negative pulse.
            _motorAndGauges.out0ClearBits (X_CLOCK);
            // Make pulse wide enough by waiting
            Thread.sleep (0, MOTOR_PULSE_NANOS);
            _motorAndGauges.out0SetBits (X_CLOCK);
        } catch (InterruptedException e) {
            return;
        }
    }

    /** Move Y one step, updating _ySteps.
     *
     * @param direction     true for increasing Y, false for decreasing
     */
    public void yStep (boolean direction)
    {
        if (direction) {
            _motorAndGauges.out0SetBits (Y_DIRECTION);
            _ySteps++;
        }
        else {
            _motorAndGauges.out0ClearBits (Y_DIRECTION);
            _ySteps--;
        }

        try {
            // negative pulse.
            _motorAndGauges.out0ClearBits (Y_CLOCK);
            // Make pulse wide enough by waiting
            Thread.sleep (0, MOTOR_PULSE_NANOS);
            _motorAndGauges.out0SetBits (Y_CLOCK);
        } catch (InterruptedException e) {
            return;
        }
    }

    /** Move probe X to the position smoothly.
     * This operates the piezoelectric directly without regard to the strain
     *   gauge readings on the probe arm.  Because of piezoelectric hysteresis,
     *   calling this method with a certain value for the position will not
     *   always result in the same actual deflection of the probe.  To get
     *   a certain deflection, use OrthoTweezersTools.seekXArmDeflection.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param position  normalized position in the range -1 to 1.
     *                  Values outside this range are clipped to this range.
     * @see OrthoTweezersTools#seekXArmDeflection
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     */
    public void moveProbeX (double position) throws CancelException {
        if (position < -1)
        position = -1;
        if (position > 1)
        position = 1;

        moveProbeHelper (position, PROBEX_CHANNEL);
        _probeXPosition = position;
    }

    /** Move probe Y to the position smoothly.
     * This operates the piezoelectric directly without regard to the strain
     *   gauge readings on the probe arm.  Because of piezoelectric hysteresis,
     *   calling this method with a certain value for the position will not
     *   always result in the same actual deflection of the probe.  To get
     *   a certain deflection, use OrthoTweezersTools.seekYArmDeflection.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param position  normalized position in the range -1 to 1.
     *                  Values outside this range are clipped to this range.
     * @see OrthoTweezersTools#seekYArmDeflection
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     */
    public void moveProbeY (double position) throws CancelException {
        if (position < -1)
        position = -1;
        if (position > 1)
        position = 1;

        // Invert position since probe Y channel is reversed.
        moveProbeHelper (-position, PROBEY_CHANNEL);
        _probeYPosition = position;
    }

    /** Return the nomalized position of probe X in the range -1 to 1.
     */
    public double probeXPosition () {
        return _probeXPosition;
    }

    /** Return the nomalized position of probe Y in the range -1 to 1.
     */
    public double probeYPosition () {
        return _probeYPosition;
    }

    /** Do the work of moveProbeX and moveProbeY.  Move the probe given by
     * probeChannel.  Assume that position has already been clipped to the
     * valid range and that probe Y has already been inverted.
     */
    private void moveProbeHelper (double position, int probeChannel) throws CancelException {
        // Convert to 12-bit integer in range 0 to 0x0fff.
        int daValueFinal = (int)(((position + 1.) / 2.) * 0x0fff);

        int increment = 10;
        int sleepMillis = 1;
        boolean increasing = (daValueFinal > _analogOutputs.lastDASet(probeChannel));
        int sign = (increasing ? 1 : -1);

        try {
            for (int daValue = _analogOutputs.lastDASet(probeChannel) + sign * increment;
            increasing ? daValue < daValueFinal : daValue > daValueFinal;
            daValue += sign * increment) {
                _analogOutputs.daSet (probeChannel, daValue);
                CancelThread.check();
                Thread.sleep(sleepMillis);
            }

            // Move the final subset of increment to the final position
            _analogOutputs.daSet (probeChannel, daValueFinal);
            CancelThread.check();
            Thread.sleep(sleepMillis);
        } catch (InterruptedException e) {
            throw new CancelException ("operation interrupted");
        }
    }

    /** Move Z one step, updating _zSteps.
     *
     * @param direction     true for increasing Z, false for decreasing
     */
    public void zStep (boolean direction)
    {
        if (direction) {
            _motorAndGauges.out1SetBits (Z_DIRECTION);
            _zSteps++;
        }
        else {
            _motorAndGauges.out1ClearBits (Z_DIRECTION);
            _zSteps--;
        }

        try {
            // negative pulse.
            _motorAndGauges.out1ClearBits (Z_CLOCK);
            // Make pulse wide enough by waiting
            Thread.sleep (0, MOTOR_PULSE_NANOS);
            _motorAndGauges.out1SetBits (Z_CLOCK);
        } catch (InterruptedException e) {
            return;
        }
    }

    /** Get the reading of the probe X arm strain gauge and add it to
     * strainOffsetXArm().
     * This performs a new A/D conversion for the strain gauge.
     *
     * @return  the strain gauge reading on the arm of probe X in volts,
     *          adjusted by the offset parameter
     * @see #strainOffsetXArm
     */
    public double xArmStrain() {
        return (readStrainGauge (PROBEX_STRAIN_ARM) + _strainOffsetXArm);
    }

    /** Get the reading of the probe Y arm strain gauge and add it to
     * strainOffsetYArm().
     * This performs a new A/D conversion for the strain gauge.
     *
     * @return  the strain gauge reading on the arm of probe Y in volts,
     *          adjusted by the offset parameter
     * @see #strainOffsetYArm
     */
    public double yArmStrain() {
        return (readStrainGauge (PROBEY_STRAIN_ARM) + _strainOffsetYArm);
    }

    /** Get the reading of the probe X tip strain gauge and add it to
     * strainOffsetXTip().
     * This performs a new A/D conversion for the strain gauge.
     *
     * @return  the strain gauge reading on the tip of probe X in volts,
     *          adjusted by the offset parameter
     * @see #strainOffsetXTip
     */
    public double xTipStrain() {
        return (readStrainGauge (PROBEX_STRAIN_TIP) + _strainOffsetXTip);
    }

    /** Get the reading of the probe Y tip strain gauge and add it to
     * strainOffsetYTip().
     * This performs a new A/D conversion for the strain gauge.
     *
     * @return  the strain gauge reading on the tip of probe Y in volts,
     *          adjusted by the offset parameter
     * @see #strainOffsetYTip
     */
    public double yTipStrain() {
        return (readStrainGauge (PROBEY_STRAIN_TIP) + _strainOffsetYTip);
    }

    /** Get the reading of the specified strain gauge.
     *
     * @param channel   the A/D channel of the strain gauge to read
     * @return  the strain gauge reading in volts
     */
    private double readStrainGauge (int channel) {
        double adValue = (double)_motorAndGauges.readAD (channel);
        // adValue is 0 to 0x0fff which represents voltages -5 to 5
        //   with 0x07ff at zero volts.
        return ((adValue - 0x07ff) * 5. / (0x0fff - 0x07ff));
    }

    /** Get the offset which is added to the probe X arm strain gauge reading.
     * You can obtain the raw strain gauge reading as
     *   xArmStrain() - strainOffsetXArm().
     *
     * @return  the probe X arm strain offset parameter in volts
     * @see #xArmStrain
     */
    public double strainOffsetXArm() {
        return _strainOffsetXArm;
    }

    /** Get the offset which is added to the probe Y arm strain gauge reading.
     * You can obtain the raw strain gauge reading as
     *   yArmStrain() - strainOffsetYArm().
     *
     * @return  the probe Y arm strain offset parameter in volts
     * @see #yArmStrain
     */
    public double strainOffsetYArm() {
        return _strainOffsetYArm;
    }

    /** Get the offset which is added to the probe X tip strain gauge reading.
     * You can obtain the raw strain gauge reading as
     *   xTipStrain() - strainOffsetXTip().
     *
     * @return  the probe X tip strain offset parameter in volts
     * @see #xTipStrain
     */
    public double strainOffsetXTip() {
        return _strainOffsetXTip;
    }

    /** Get the offset which is added to the probe Y tip strain gauge reading.
     * You can obtain the raw strain gauge reading as
     *   yTipStrain() - strainOffsetYTip().
     *
     * @return  the probe Y tip strain offset parameter in volts
     * @see #yTipStrain
     */
    public double strainOffsetYTip() {
        return _strainOffsetYTip;
    }

    /** Get the parameter to convert strain to deflection on the probe X arm.
     * The deflection in microns is xArmStrain()*strainToDeflectionXArm().
     *
     * @return  the probe X arm strain to deflection conversion parameter
     *          from volts to millimeters
     */
    public double strainToDeflectionXArm() {
        return _strainToDeflectionXArm;
    }

    /** Get the parameter to convert strain to deflection on the probe X tip.
     * The deflection in microns is xTipStrain()*strainToDeflectionXTip().
     *
     * @return  the probe X tip strain to deflection conversion parameter
     *          from volts to millimeters
     */
    public double strainToDeflectionXTip() {
        return _strainToDeflectionXTip;
    }

    /** Get the parameter to convert strain to deflection on the probe Y arm.
     * The deflection in microns is yArmStrain()*strainToDeflectionYArm().
     *
     * @return  the probe Y arm strain to deflection conversion parameter
     *          from volts to millimeters
     */
    public double strainToDeflectionYArm() {
        return _strainToDeflectionYArm;
    }

    /** Get the parameter to convert strain to deflection on the probe Y tip.
     * The deflection in microns is yTipStrain()*strainToDeflectionYTip().
     *
     * @return  the probe Y tip strain to deflection conversion parameter
     *          from volts to millimeters
     */
    public double strainToDeflectionYTip() {
        return _strainToDeflectionYTip;
    }

    /** Get the parameter to convert strain to force on the probe X tip.
     * The force in milli Newtons is xTipStrain()*strainToForceXTip().
     *
     * @return  the probe X tip strain to force conversion parameter
     *          from volts to milli Netwons
     */
    public double strainToForceXTip() {
        return _strainToForceXTip;
    }

    /** Get the parameter to convert strain to force on the probe Y tip.
     * The force in milli Newtons is yTipStrain()*strainToForceYTip().
     *
     * @return  the probe Y tip strain to force conversion parameter
     *          from volts to milli Netwons
     */
    public double strainToForceYTip() {
        return _strainToForceYTip;
    }

    /** Get the deflection of the probe X arm.
     * This performs a new A/D conversion for the strain gauge
     *
     * @return  the deflection in millimeters
     */
    public double xArmDeflection() {
        return (xArmStrain() * _strainToDeflectionXArm);
    }

    /** Get the deflection of the probe X tip.
     * This performs a new A/D conversion for the strain gauge.
     * Note that xTipForce() also reads the probe X tip strain gauge.
     * If you only want to read the gauge once to get them both, you can call
     * xTipStrain() and multiply by strainToDeflectionXTip() and
     * strainToForceXTip().
     *
     * @return  the deflection in millimeters
     * @see     #xTipStrain
     * @see     #strainToDeflectionXTip
     * @see     #strainToForceXTip
     */
    public double xTipDeflection() {
        return (xTipStrain() * _strainToDeflectionXTip);
    }

    /** Get the deflection of the probe Y arm.
     * This performs a new A/D conversion for the strain gauge
     *
     * @return  the deflection in millimeters
     */
    public double yArmDeflection() {
        return (yArmStrain() * _strainToDeflectionYArm);
    }

    /** Get the deflection of the probe Y tip.
     * This performs a new A/D conversion for the strain gauge.
     * Note that yTipForce() also reads the probe Y tip strain gauge.
     * If you only want to read the gauge once to get them both, you can call
     * yTipStrain() and multiply by strainToDeflectionYTip() and
     * strainToForceXYTip().
     *
     * @return  the deflection in millimeters
     * @see     #yTipStrain
     * @see     #strainToDeflectionYTip
     * @see     #strainToForceYTip
     */
    public double yTipDeflection() {
        return (yTipStrain() * _strainToDeflectionYTip);
    }

    /** Get the force on the probe X tip.
     * This performs a new A/D conversion for the strain gauge.
     * Note that xTipDeflection() also reads the probe X tip strain gauge.
     * If you only want to read the gauge once to get them both, you can call
     * xTipStrain() and multiply by strainToDeflectionXTip() and
     * strainToForceXTip().
     *
     * @return  the force in milli Newtons
     * @see     #xTipStrain
     * @see     #strainToDeflectionXTip
     * @see     #strainToForceXTip
     */
    public double xTipForce() {
        return (xTipStrain() * _strainToForceXTip);
    }

    /** Get the force on the probe Y tip.
     * This performs a new A/D conversion for the strain gauge.
     * Note that yTipDeflection() also reads the probe Y tip strain gauge.
     * If you only want to read the gauge once to get them both, you can call
     * yTipStrain() and multiply by strainToDeflectionYTip() and
     * strainToForceYTip().
     *
     * @return  the force in milli Newtons
     * @see     #yTipStrain
     * @see     #strainToDeflectionYTip
     * @see     #strainToForceYTip
     */
    public double yTipForce() {
        return (yTipStrain() * _strainToForceYTip);
    }

    /** Move the stage to the x, y position, moving diagonally if necessary.
     * Note that the coordinates are as if there are axes painted on the stage,
     * so that the probes reach that position.
     * That is, for the probes to get to increasing X, the stage itself actually
     * moves in the negative direction.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param x     x coordinate to move to, in millimeters
     * @param y     y coordinate to move to, in millimeters
     * @exception CancelException   thrown by CancelThread.check()
     *              or if got an InterruptedException.
     * @see         CancelThread
     */
    public void moveToXY (double x, double y) throws CancelException {
        int xStepsFinal = (int)(x * STEPS_PER_MM);
        int yStepsFinal = (int)(y * STEPS_PER_MM);
        int deltaX = xStepsFinal - _xSteps;
        int deltaY = yStepsFinal - _ySteps;
        if (deltaX == 0 && deltaY == 0) {
            // Already at the desired position
            return;
        }

        boolean deltaXBigger = (Math.abs (deltaX) > Math.abs (deltaY));
        boolean xDirection = (deltaX > 0);
        boolean yDirection = (deltaY > 0);

        try {
            if (deltaXBigger) {
                // Step x and occasionally step y.
                double slope = (double)deltaY / (double)deltaX;
                while (_xSteps != xStepsFinal) {
                    CancelThread.check();

                    xStep (xDirection);
                    double expectedY = (double)yStepsFinal + (_xSteps - xStepsFinal)*slope;
                    if (yDirection) {
                        // Increasing y
                        if (expectedY - _ySteps >= 1.)
                        yStep (true);
                    }
                    else {
                        // Decreasing y
                        if (_ySteps - expectedY >= 1.)
                        yStep (false);
                    }

                    Thread.sleep (STAGE_STEP_MILLIS, STAGE_STEP_NANOS);
                }

                // _xSteps is now the correct value, but _ySteps may be a little off.
                // Recompute yDirection and move as necessary
                yDirection = (yStepsFinal > _ySteps);
                while (_ySteps != yStepsFinal) {
                    CancelThread.check();

                    yStep (yDirection);
                    Thread.sleep (STAGE_STEP_MILLIS, STAGE_STEP_NANOS);
                }
            }
            else {
                // delta Y is bigger
                // Step y and occasionally step x.
                double slope = (double)deltaX / (double)deltaY;
                while (_ySteps != yStepsFinal) {
                    CancelThread.check();

                    yStep (yDirection);
                    double expectedX = (double)xStepsFinal + (_ySteps - yStepsFinal)*slope;
                    if (xDirection) {
                        // Increasing x
                        if (expectedX - _xSteps >= 1.)
                        xStep (true);
                    }
                    else {
                        // Decreasing x
                        if (_xSteps - expectedX >= 1.)
                        xStep (false);
                    }

                    Thread.sleep (STAGE_STEP_MILLIS, STAGE_STEP_NANOS);
                }

                // _ySteps is now the correct value, but _xSteps may be a little off.
                // Recompute xDirection and move as necessary
                xDirection = (xStepsFinal > _xSteps);
                while (_xSteps != xStepsFinal) {
                    CancelThread.check();

                    xStep (xDirection);
                    Thread.sleep (STAGE_STEP_MILLIS, STAGE_STEP_NANOS);
                }
            }
        } catch (InterruptedException e) {
            throw new CancelException (e.getMessage());
        }
    }

    /** Move the stage to the z position.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param z     z coordinate to move to, in millimeters
     * @exception CancelException   thrown by CancelThread.check()
     *              or if got an InterruptedException.
     * @see         CancelThread
     */
    public void moveToZ (double z) throws CancelException {
        int zStepsFinal = (int)(z * STEPS_PER_MM);
        boolean direction = (zStepsFinal > _zSteps);

        try {
            while (_zSteps != zStepsFinal) {
                CancelThread.check();

                zStep (direction);
                Thread.sleep (STAGE_STEP_MILLIS, STAGE_STEP_NANOS);
            }
        } catch (InterruptedException e) {
            throw new CancelException (e.getMessage());
        }
    }

    /** Return the X position of the stage in millimeters.
     */
    public double stageX () {
        return (_xSteps / STEPS_PER_MM);
    }
    /** Return the Y position of the stage in millimeters.
     */
    public double stageY () {
        return (_ySteps / STEPS_PER_MM);
    }
    /** Return the Z position of the stage in millimeters.
     */
    public double stageZ () {
        return (_zSteps / STEPS_PER_MM);
    }

    /** Set the present position of the stage.  This can be used
     * when the application starts to return the state to what it was
     * during the previous session.  Or, all values of zero can be used
     * to reset the origin.
     *
     * @param x     the new stage X position in millimeters
     * @param y     the new stage Y position in millimeters
     * @param z     the new stage Z position in millimeters
     */
    public void setStageXYZ (double x, double y, double z) {
        _xSteps = (int)(x * STEPS_PER_MM);
        _ySteps = (int)(y * STEPS_PER_MM);
        _zSteps = (int)(z * STEPS_PER_MM);
    }

    /** Assume there is no force on the tip of probe X and set
     * strainOffsetXTip so that xTipStrain() will read zero.
     *
     * @see #xTipStrain
     */
    public void resetXTipStrain() {
        // Get the raw readings and set the new offset to its negative
        _strainOffsetXTip = -(xTipStrain() - _strainOffsetXTip);
    }

    /** Assume there is no force on the tip of probe Y and set
     * strainOffsetYTip so that yTipStrain() will read zero.
     *
     * @see #yTipStrain
     */
    public void resetYTipStrain() {
        // Get the raw readings and set the new offset to its negative
        _strainOffsetYTip = -(yTipStrain() - _strainOffsetYTip);
    }

    /** Assume that probe X is at the zero position and set
     * strainOffsetXArm so that xArmStrain() will read zero.
     *
     * @see #xArmStrain
     */
    public void resetXArmStrain() {
        // Get the raw readings and set the new offset to its negative
        _strainOffsetXArm = -(xArmStrain() - _strainOffsetXArm);
    }

    /** Assume that probe Y is at the zero position and set
     * strainOffsetYArm so that yArmStrain() will read zero.
     *
     * @see #yArmStrain
     */
    public void resetYArmStrain() {
        // Get the raw readings and set the new offset to its negative
        _strainOffsetYArm = -(yArmStrain() - _strainOffsetYArm);
    }

    /** This adjusts strainToDeflectionXArm so that the current reading of
     * xArmStrain() will yeild the given deflection.  This assumes that
     * strainOffsetXArm has already been calibrated using resetXArmStrain().
     *
     * @param deflection  the actual deflection in millimeters that probe X
     *                  arm is believed to be at
     * @see #resetXArmStrain
     */
    public void calibrateXArmDeflection (double deflection) {
        // Assume that xArmStrain() is not zero.
        _strainToDeflectionXArm = deflection / xArmStrain();
    }

    /** This adjusts strainToDeflectionYArm so that the current reading of
     * yArmStrain() will yeild the given deflection.  This assumes that
     * strainOffsetYArm has already been calibrated using resetYArmStrain().
     *
     * @param deflection  the actual deflection in millimeters that probe Y
     *                  arm is believed to be at
     * @see #resetYArmStrain
     */
    public void calibrateYArmDeflection (double deflection) {
        // Assume that yArmStrain() is not zero.
        _strainToDeflectionYArm = deflection / yArmStrain();
    }

    /** This adjusts strainToDeflectionXTip so that the current reading of
     * xTipStrain() will yeild the given deflection.  This assumes that
     * strainOffsetXTip has already been calibrated using resetXTipStrain().
     *
     * @param deflection  the actual deflection in millimeters that probe X
     *                  tip is believed to be at
     * @see #resetXTipStrain
     */
    public void calibrateXTipDeflection (double deflection) {
        // Assume that xTipStrain() is not zero.
        _strainToDeflectionXTip = deflection / xTipStrain();
    }

    /** This adjusts strainToDeflectionYTip so that the current reading of
     * yTipStrain() will yeild the given deflection.  This assumes that
     * strainOffsetYTip has already been calibrated using resetYTipStrain().
     *
     * @param deflection  the actual deflection in millimeters that probe Y
     *                  tip is believed to be at
     * @see #resetYTipStrain
     */
    public void calibrateYTipDeflection (double deflection) {
        // Assume that yTipStrain() is not zero.
        _strainToDeflectionYTip = deflection / yTipStrain();
    }

    /** This adjusts strainToForceXTip so that the current reading of
     * xTipStrain() will yeild the given force.  This assumes that
     * strainOffsetXTip has already been calibrated using resetXTipStrain().
     *
     * @param force  the actual force in milli Newtons that probe X
     *                  tip is believed to be at
     * @see #resetXTipStrain
     */
    public void calibrateXTipForce (double force) {
        // Assume that xTipStrain() is not zero.
        _strainToForceXTip = force / xTipStrain();
    }

    /** This adjusts strainToForceYTip so that the current reading of
     * yTipStrain() will yeild the given force.  This assumes that
     * strainOffsetYTip has already been calibrated using resetYTipStrain().
     *
     * @param force  the actual force in milli Newtons that probe Y
     *                  tip is believed to be at
     * @see #resetYTipStrain
     */
    public void calibrateYTipForce (double force) {
        // Assume that yTipStrain() is not zero.
        _strainToForceYTip = force / yTipStrain();
    }

    /** This sets xArmStrainError, xTipStrainError, yArmStrainError and
     * yTipStrain error by running a calibration algorithm.  The calibration
     * algorithm reads the each strain gauge for several iterations, computing
     * the maximum and minimum readings.  The error is maximum - minimum.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     */
    public void calibrateStrainErrors () throws CancelException {
        int iterations = 1000;
        double min, max;

        min = Double.MAX_VALUE;
        max = -Double.MAX_VALUE;
        for (int i = 0; i < iterations; ++i) {
            double value = xArmStrain();
            min = Math.min (min, value);
            max = Math.max (max, value);

            CancelThread.check();
        }
        _strainErrorXArm = max - min;

        min = Double.MAX_VALUE;
        max = -Double.MAX_VALUE;
        for (int i = 0; i < iterations; ++i) {
            double value = xTipStrain();
            min = Math.min (min, value);
            max = Math.max (max, value);

            CancelThread.check();
        }
        _strainErrorXTip = max - min;

        min = Double.MAX_VALUE;
        max = -Double.MAX_VALUE;
        for (int i = 0; i < iterations; ++i) {
            double value = yArmStrain();
            min = Math.min (min, value);
            max = Math.max (max, value);

            CancelThread.check();
        }
        _strainErrorYArm = max - min;

        min = Double.MAX_VALUE;
        max = -Double.MAX_VALUE;
        for (int i = 0; i < iterations; ++i) {
            double value = yTipStrain();
            min = Math.min (min, value);
            max = Math.max (max, value);

            CancelThread.check();
        }
        _strainErrorYTip = max - min;
    }

    public double strainErrorXArm() {
        return _strainErrorXArm;
    }

    public double strainErrorXTip() {
        return _strainErrorXTip;
    }

    public double strainErrorYArm() {
        return _strainErrorYArm;
    }

    public double strainErrorYTip() {
        return _strainErrorYTip;
    }

    /** Set the digital output for relay number 1 to 1.  You may check the state with
     * relay1State()
     *
     * @see #relay1State
     */
    public void relay1On() {
        _motorAndGauges.out1SetBits(RELAY_1);
        _relay1State = true;
    }
    
    /** Set the digital output for relay number 1 to 0.  You may check the state with
     * relay1State()
     *
     * @see #relay1State
     */
    public void relay1Off() {
        _motorAndGauges.out1ClearBits(RELAY_1);
        _relay1State = false;
    }
    
    /** Return true if relay number 1 is on, otherwise false.
     *
     * @see #relay1On
     * @see #relay1Off
     */
    public boolean relay1State() {
        return _relay1State;
    }
    
    // bit masks for _motorAndGauges digital output 0
    private final static int X_CLOCK = 0x01;
    private final static int X_DIRECTION = 0x02;
    private final static int Y_CLOCK = 0x10;
    private final static int Y_DIRECTION = 0x20;
    // bit masks for _motorAndGauges digital output 1
    private final static int Z_CLOCK = 0x01;
    private final static int Z_DIRECTION = 0x02;
    private final static int UV_GUN = 0x10;
    private final static int RELAY_1 = 0x20;
    // Analog output channel numbers
    private final static int PROBEX_CHANNEL = 1;
    private final static int PROBEY_CHANNEL = 2;
    // Analog input channel numbers
    private final static int PROBEX_STRAIN_ARM = 0;
    private final static int PROBEX_STRAIN_TIP = 1;
    private final static int PROBEY_STRAIN_ARM = 2;
    private final static int PROBEY_STRAIN_TIP = 3;

    // Numer of milli and nanoseconds to sleep between stepping the stage motors
    private final static long STAGE_STEP_MILLIS = 0L;
    private final static int STAGE_STEP_NANOS = 500000;
    // nanosecond width of pulse to step the stage motors
    private final static int MOTOR_PULSE_NANOS = 10;
    // Stage steps per millimeter
    private final static double STEPS_PER_MM = 1666.6;

    private AdvantechPCL711B _motorAndGauges;
    private AdvantechPCL726 _analogOutputs;
    private int _xSteps, _ySteps, _zSteps;
    private double _probeXPosition, _probeYPosition;
    private boolean _relay1State;

    // Strain reading error which is max - min over several iterations
    private double _strainErrorXArm = .04, _strainErrorXTip = .04;
    private double _strainErrorYArm = .04, _strainErrorYTip = .04;

    // Zero offset for strain gauge readings
    private double _strainOffsetXArm = 0, _strainOffsetYArm = 0;
    private double _strainOffsetXTip = 0, _strainOffsetYTip = 0;

    // Converts volts to microns of deflection. Initilize to defaults.
    private double _strainToDeflectionXArm = -0.2, _strainToDeflectionYArm = -0.2;
    private double _strainToDeflectionXTip = .01, _strainToDeflectionYTip = .01;

    // Converts volts to milliNewtons of force. Initilize to defaults.
    private double _strainToForceXTip = 1, _strainToForceYTip = 1;
}
