/*
Copyright (c) 2000,2001. The Regents of the University of California
(Regents). All Rights Reserved. 

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research, and not-for-profit purposes,
without fee and without a signed licensing agreement, is hereby
granted, provided that the above copyright notice, this paragraph and
the following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC
Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620,
(510) 643-7201, for commercial licensing opportunities.
  
IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. 
  
REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF
ANY, PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO
OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.
*/
				 
package orthotweezers;

import javax.vecmath.*;

/** OrthoTweezersTools is a utility class for doing more sophisticated operations
 * than are provided by the basic OrthoTweezersControl class.  Logically, high level
 * manipulation routines don't belong in the basic OrthTweezersControl class which would just
 * clutter it.  Note that an OrthoTweezersTools object does not keep any local
 * variables (except for the OrthoTweezersControl object passed to the constructor).  
 * It is just a bunch of utility methods.  Therefore, any code can create an
 * OrthoTweezersTools object from an OrthoTweezersControl whenever it needs the
 * utility routines.
 *
 * @see OrthoTweezersControl
 *
 * @author  unknown
 */
public class OrthoTweezersTools extends Object {

    /** Creates new OrthoTweezersTools which uses the given control object.
     *
     * @param control   The OrthoTweezersControl object which the methods will use. 
     */
    public OrthoTweezersTools (OrthoTweezersControl control) {
        _control = control;
    }
    
    /** Returns the OrthoTweezersControl object which was passed to the 
     * constructor.
     */
    public OrthoTweezersControl control() {
        return _control;
    }

    /** Set the point to the coordinates of the stage.  To return a new Point3d
     * object with the coordinates, use stagePoint().
     *
     * @param point     the object which will receive the stage coordiates, in millimeters
     * @see #stagePoint
     * @see OrthoTweezersControl#stageX
     */
    public void getStageXYZ (Point3d point) {
        point.set (_control.stageX(), _control.stageY(), _control.stageZ());
    }
    
    /** Return a new Point3d containing the stage X,Y,Z position in millimeters.  To set an
     * existing Point3d object, use getStageXYZ(point).
     *
     * @see #getStageXYZ
     */
    public Point3d stagePoint() {
        return new Point3d (_control.stageX(), _control.stageY(), _control.stageZ());
    }
    
    /** Move the stage in X and Y directions relative to the current position.
     * This method is identical to
     * control().moveToXY(control().stageX() + deltaX, control().stageY() + deltaY).
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param deltaX     the X distance to move, in millimeters
     * @param deltaY     the Y distance to move, in millimeters
     * @exception CancelException   thrown by CancelThread.check()
     *
     * @see OrthoTweezersControl#moveToXY
     * @see         CancelThread
     */
    public void moveDeltaXY(double deltaX, double deltaY) throws CancelException {
        _control.moveToXY(_control.stageX() + deltaX, _control.stageY() + deltaY);
    }
    
    /** Move the stage in the Z direction relative to the current position.
     * This method is identical to
     * control().moveToZ(control().stageZ() + deltaZ).
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param deltaZ     the Z distance to move, in millimeters
     * @exception CancelException   thrown by CancelThread.check()
     *
     * @see OrthoTweezersControl#moveToZ
     * @see         CancelThread
     */
    public void moveDeltaZ(double deltaZ) throws CancelException {
        _control.moveToZ(_control.stageZ() + deltaZ);
    }
    
    /** Move the stage to the XYZ coordinates.  If the Z coordinate to move
     * to is higher than the presens Z value, this moves to the Z coordinate first, otherwise
     * this moves to the XY coordinates first, then to the Z coordinate.  This
     * is to make sure the XY motion happens at the highest Z value to avoid crashing
     * into objects on the surface.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param x     x coordinate to move to, in millimeters
     * @param y     y coordinate to move to, in millimeters
     * @param z     z coordinate to move to, in millimeters
     * @exception CancelException   thrown by CancelThread.check()
     *
     * @see OrthoTweezersControl#moveToXY
     * @see OrthoTweezersControl#moveToZ
     * @see         CancelThread
     */
    public void moveToXYZ(double x, double y, double z) throws CancelException {
        if (z > _control.stageZ()) {
            _control.moveToZ (z);
            _control.moveToXY (x, y);
        }
        else {
            _control.moveToXY (x, y);
            _control.moveToZ (z);
        }
    }
    
    /** This just calls moveToXYZ(x,y,z) with the coordinates of the point.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param point     the point containing the coordinates to move to, in millimeters
     * @exception CancelException   thrown by CancelThread.check()
     *
     * @see #moveToXYZ
     * @see         CancelThread
     */
    public void moveToPoint (Point3d point) throws CancelException {
        moveToXYZ (point.x, point.y, point.z);
    }
    
    /** Move the X probe relative to the current position.
     * This method is identical to
     * control().moveProbeX(control().probeXPosition() + deltaPosition).
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param deltaPosition     the distance to move the probe X position, in the
     *                  same normalized units from -1 to 1 that moveProbeX uses
     * @exception CancelException   thrown by CancelThread.check()
     *
     * @see OrthoTweezersControl#moveProbeX
     * @see         CancelThread
     */
    public void moveProbeDeltaX(double deltaPosition) throws CancelException {
        _control.moveProbeX(_control.probeXPosition() + deltaPosition);
    }
    
    /** Move the Y probe relative to the current position.
     * This method is identical to
     * control().moveProbeY(control().probeYPosition() + deltaPosition).
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param deltaPosition     the distance to move the probe Y position, in the
     *                  same normalized units from -1 to 1 that moveProbeY uses
     * @exception CancelException   thrown by CancelThread.check()
     *
     * @see OrthoTweezersControl#moveProbeY
     * @see         CancelThread
     */
    public void moveProbeDeltaY(double deltaPosition) throws CancelException {
        _control.moveProbeY(_control.probeYPosition() + deltaPosition);
    }
    
    /** Move probe X until its arm deflection is the given value.  This uses
     * feedback from the arm (but not tip) strain gauge.  To move the probe directly
     * without any feedback, use moveProbeX.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param deflection    the deflection to seek to in millimeters
     * @exception CancelException   thrown by CancelThread.check()
     *              or if got an InterruptedException.
     * @see         CancelThread
     * @see OrthoTweezersControl#moveProbeX
     */
    public void seekXArmDeflection (double deflection) throws CancelException {
        try {
            // Do this whole algorithm twice, the first time to get in the
            //   general neighborhood, then wait a little for the probe dynamics
            //   to settle down and then seek again to get closer.
            for (int i = 0; i < 2; ++i) {
                boolean increasing = (_control.xArmDeflection() < deflection);
                double sign = (increasing ? 1. : -1);
                // The first time through, step by more than SEEK_POSITION_INCREMENT to go faster
                for (double position = _control.probeXPosition();
                     increasing ? position < 1. && _control.xArmDeflection() < deflection
                                : position > -1. && _control.xArmDeflection() > deflection;
                     position += sign * (i == 0 ? 3 : 1) * SEEK_POSITION_INCREMENT) {
                    _control.moveProbeX (position);

                    CancelThread.check();
                }

                if (i == 0)
                  Thread.sleep (ARM_SETTLING_MILLIS);
            }
        } catch (InterruptedException e) {
            throw new CancelException ("interrupted exception");
        }
    }

    /** Move probe Y until its arm deflection is the given value.  This uses
     * feedback from the arm (but not tip) strain gauge.  To move the probe directly
     * without any feedback, use moveProbeY.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param deflection    the deflection to seek to in millimeters
     * @exception CancelException   thrown by CancelThread.check()
     *              or if got an InterruptedException.
     * @see         CancelThread
     * @see OrthoTweezersControl#moveProbeY
     */
    public void seekYArmDeflection (double deflection) throws CancelException {
        try {
            // Do this whole algorithm twice, the first time to get in the
            //   general neighborhood, then wait a little for the probe arm dynamics
            //   to settle down and then seek again to get closer.
            for (int i = 0; i < 2; ++i) {
                boolean increasing = (_control.yArmDeflection() < deflection);
                double sign = (increasing ? 1. : -1);
                // The first time through, step by more than SEEK_POSITION_INCREMENT to go faster
                for (double position = _control.probeYPosition();
                     increasing ? position < 1. && _control.yArmDeflection() < deflection
                                : position > -1. && _control.yArmDeflection() > deflection;
                     position += sign * (i == 0 ? 3 : 1) * SEEK_POSITION_INCREMENT) {
                    _control.moveProbeY (position);

                    CancelThread.check();
                }

                if (i == 0)
                  Thread.sleep (ARM_SETTLING_MILLIS);
            }
        } catch (InterruptedException e) {
            throw new CancelException ("interrupted exception");
        }
    }

    /** This uses seekXArmDeflection and seekYArmDeflection to seek to the
     * given X and Y deflections.  If the Y arm deflection is increasing,
     * this moves the Y arm to its deflection first, otherwise, this moves
     * the X arm first.  This is to prevent the case where probe X is decreasing
     * while probe Y is increasing, but probe X is moved first and could crash
     * into probe Y.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param xDeflection    the deflection to seek for the X arm in millimeters
     * @param yDeflection    the deflection to seek for the Y arm in millimeters
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     * @see         #seekXArmDeflection
     * @see         #seekYArmDeflection
     */
    public void seekXYArmDeflection (double xDeflection, double yDeflection) 
      throws CancelException {
          if (yDeflection > _control.yArmDeflection()) {
              seekYArmDeflection (yDeflection);
              seekXArmDeflection (xDeflection);
          }
          else {
              seekXArmDeflection (xDeflection);
              seekYArmDeflection (yDeflection);
          }
    }
    
    /** This resets the strain offsets for probe X and probe Y tips so that
     * the strain will now read zero.  This just calls control().resetXTipStrain()
     * and control().resetYTipStrain().  Because the strain gauge readings
     * drift over time, you should call this whenever you are sure that the
     * tips are not touching anything.
     */
    public void resetTipStrains() {
        _control.resetXTipStrain();
        _control.resetYTipStrain();
    }
    
    /** Seek the X probe arm deflection relative to the current deflection.
     * This method is identical to
     * seekXArmDeflection(control().xArmDeflection() + deltaDeflection).
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param deltaDeflection     the distance to deflect the probe X arm in
     *                      millimeters relative to the current deflection.
     * @exception CancelException   thrown by CancelThread.check()
     *
     * @see #seekXArmDeflection
     * @see         CancelThread
     */
    public void seekXArmDeltaDeflection(double deltaDeflection) throws CancelException {
        seekXArmDeflection(_control.xArmDeflection() + deltaDeflection);
    }
    
    /** Seek the Y probe arm deflection relative to the current deflection.
     * This method is identical to
     * seekYArmDeflection(control().yArmDeflection() + deltaDeflection).
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param deltaDeflection     the distance to deflect the probe Y arm in
     *                      millimeters relative to the current deflection.
     * @exception CancelException   thrown by CancelThread.check()
     *
     * @see #seekYArmDeflection
     * @see         CancelThread
     */
    public void seekYArmDeltaDeflection(double deltaDeflection) throws CancelException {
        seekYArmDeflection(_control.yArmDeflection() + deltaDeflection);
    }
    
    /** Move probe X until the tip strain gauge detects a contact force,
     * then back off a little until the force drops near zero.  If increasing
     * is true then the probe position increases, otherwise decreases.
     * To do a similar
     * operation by moving the stage instead of the probe, use moveStageXToContact.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param increasing    true to increase the probe position to contact,
     *                      false to decrease
     * @return  true for success, false if couldn't find contact.  Note that this
     *      may return false if the probe is already contacting something, but in
     *      the opposite direction as indicated by increasing.
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     * @see     #moveStageXToContact
     * @see     #moveProbeXToContactBounded
     */
    public boolean moveProbeXToContact (boolean increasing) throws CancelException {
        // Use rediculously large values for minDeflection and maxDeflection
        // so that this will limit to the min and max positions of -1 or 1.
        return moveProbeXToContactBounded (increasing, -1000., 1000.);
    }

    /** Move probe X, bounded by the given range, until the tip strain gauge detects a contact force,
     * then back off a little until the force drops near zero.  If increasing
     * is true then the probe position increases, otherwise decreases.
     * To do a similar
     * operation by moving the stage instead of the probe, use moveStageXToContact.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param increasing    true to increase the probe position to contact,
     *                      false to decrease
     * @param minDeflection minimum probe X deflection (in millimeters) allowed 
     *                  when seeking the contact. This also stops if 
     *                  control.probeXPosition() reaches the absolute limit of -1.
     *                  If minDeflection >= maxDeflection, this simply returns false.
     * @param maxDeflection maximum probe X deflection (in millimeters) allowed 
     *                  when seeking the contact. This also stops if 
     *                  control.probeXPosition() reaches the absolute limit of +1.
     *                  If maxDeflection <= minDeflection, this simply returns false.
     * @return  true for success, false if couldn't find contact.  Note that this
     *      may return false if the probe is already contacting something, but in
     *      the opposite direction as indicated by increasing.
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     * @see     #moveStageXToContact
     * @see     #moveProbeXToContact
     */
    public boolean moveProbeXToContactBounded 
       (boolean increasing, double minDeflection, double maxDeflection) throws CancelException {
        double sign = (increasing ? 1. : -1);        
        // Touch with a force 3 times the strain error (converted to force).
        // For increasing probe position, the force is negative
        // Note that it doesn't matter what _strainToForceXTip is, as long as it
        //   is not zero, because seekXTipForce uses the same value, so this
        //   really only reads the strain gauge directly.
        if (!seekXTipForceBounded (-sign * 3. * _control.strainToForceXTip() * _control.strainErrorXTip(),
                                   minDeflection, maxDeflection))
            return false;
        // Now, release.
        if (!seekXTipForceBounded (0., minDeflection, maxDeflection))
            return false;
        
        return true;
    }

    /** Close probe Y until the tip strain gauge detects a contact force,
     * then back off a little until the force drops near zero.  If increasing
     * is true then the probe position increases, otherwise decreases.
     * To do a similar
     * operation by moving the stage instead of the probe, use moveStageYToContact.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param increasing    true to increase the probe position to contact,
     *                      false to decrease
     * @return  true for success, false if couldn't find contact.  Note that this
     *      may return false if the probe is already contacting something, but in
     *      the opposite direction as indicated by increasing.
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     * @see         #moveStageYToContact
     * @see     #moveProbeYToContactBounded
     */
    public boolean moveProbeYToContact (boolean increasing) throws CancelException {
        // Use rediculously large values for minDeflection and maxDeflection
        // so that this will limit to the min and max positions of -1 or 1.
        return moveProbeYToContactBounded (increasing, -1000., 1000.);
    }

    /** Move probe Y, bounded by the given range, until the tip strain gauge detects a contact force,
     * then back off a little until the force drops near zero.  If increasing
     * is true then the probe position increases, otherwise decreases.
     * To do a similar
     * operation by moving the stage instead of the probe, use moveStageXToContact.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param increasing    true to increase the probe position to contact,
     *                      false to decrease
     * @param minDeflection minimum probe Y deflection (in millimeters) allowed 
     *                  when seeking the contact. This also stops if 
     *                  control.probeYPosition() reaches the absolute limit of -1.
     *                  If minDeflection >= maxDeflection, this simply returns false.
     * @param maxDeflection maximum probe Y deflection (in millimeters) allowed 
     *                  when seeking the contact. This also stops if 
     *                  control.probeYPosition() reaches the absolute limit of +1.
     *                  If maxDeflection <= minDeflection, this simply returns false.
     * @return  true for success, false if couldn't find contact.  Note that this
     *      may return false if the probe is already contacting something, but in
     *      the opposite direction as indicated by increasing.
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     * @see     #moveStageXToContact
     * @see     #moveProbeXToContact
     */
    public boolean moveProbeYToContactBounded 
       (boolean increasing, double minDeflection, double maxDeflection) throws CancelException {
        double sign = (increasing ? 1. : -1);        
        // Touch with a force 3 times the strain error (converted to force).
        // For increasing probe position, the force is negative
        // Note that it doesn't matter what _strainToForceYTip is, as long as it
        //   is not zero, because seekYTipForce uses the same value, so this
        //   really only reads the strain gauge directly.
        if (!seekYTipForceBounded (-sign * 3. * _control.strainToForceYTip() * _control.strainErrorYTip(),
                                   minDeflection, maxDeflection))
            return false;
        // Now, release.
        if (!seekYTipForceBounded (0., minDeflection, maxDeflection))
            return false;
        
        return true;
    }

    /** This calls moveProbeXToContact repeatedly, moving SWEEP_TO_CONTACT_INCREMENT
     * millimeters in the
     * Y direction until contact, or until the maxDeltaY has been reached.
     * In this way, it will sweep the X probe until it contacts an object, in which
     * case this returns true.  NOTE: This assumes that the probe Y position is
     * greater than zero such that the moving of probe X will not contact it.
     * Upon success, you can determine the position of the
     * contacted object as stageX() + xArmDeflection().
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param increasing    true to increase the probe position to contact,
     *                      false to decrease. If increasing is true, this
     *              initially moves probe X position to -1, otherwise it initially
     *              moves probe X position to +1.  This assumes the path is clear
     *              to move the probe to the initial position.  Also, once the probe is
     *              move to the initial position, this calls control().resetXTipStrain()
     *              to recalibrate the zero offset of the probe so that it is sensitive.
     * @param maxDeltaY     the maximun distance to travel in increments of
     *              SWEEP_TO_CONTACT_INCREMENT while trying to get moveProbeXToContact
     *              to return true.  The maxDeltaY is positive since we only sweep
     *              while moving in the positive Y direction.
     * @return  true for success, false if couldn't find contact.
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     * @see     #moveProbeXToContact
     */
    public boolean sweepProbeXToContact (boolean increasing, double maxDeltaY) 
    throws CancelException {
        double saveX = _control.stageX();
        double saveY = _control.stageY();

        // Move to initial position and reset strain so that reading is sensitive.
        _control.moveProbeX (increasing ? -1. : 1.);
        _control.resetXTipStrain();
        
        for (double y = saveY; y < saveY + maxDeltaY; y += SWEEP_TO_CONTACT_INCREMENT) {
            // Prepare probe position for before we move
            _control.moveProbeX (increasing ? -1. : 1.);

            _control.moveToXY (saveX, y);
            if (moveProbeXToContact (increasing))
                // contacted the block
                return (true);

            CancelThread.check();
        }
        
        return (false);
    }
    
    /** This calls moveProbeYToContact repeatedly, moving SWEEP_TO_CONTACT_INCREMENT
     * millimeters in the
     * X direction until contact, or until the maxDeltaX has been reached.
     * In this way, it will sweep the Y probe until it contacts an object, in which
     * case this returns true.  NOTE: This assumes that the probe X position is
     * greater than zero such that the moving of probe Y  will not contact it.
     * Upon success, you can determine the position of the
     * contacted object as stageY() + yArmDeflection().
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param increasing    true to increase the probe position to contact,
     *                      false to decrease. If increasing is true, this
     *              initially moves probe Y position to -1, otherwise it initially
     *              moves probe Y position to +1.  This assumes the path is clear
     *              to move the probe to the initial position.  Also, once the probe is
     *              move to the initial position, this calls control().resetYTipStrain()
     *              to recalibrate the zero offset of the probe so that it is sensitive.
     * @param maxDeltaY     the maximun distance to travel in increments of
     *              SWEEP_TO_CONTACT_INCREMENT while trying to get moveProbeYToContact
     *              to return true.  The maxDeltaX is positive since we only sweep
     *              while moving in the positive X direction.
     * @return  true for success, false if couldn't find contact.
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     * @see     #moveProbeYToContact
     */
    public boolean sweepProbeYToContact (boolean increasing, double maxDeltaX) 
    throws CancelException {
        double saveX = _control.stageX();
        double saveY = _control.stageY();

        // Move to initial position and reset strain so that reading is sensitive.
        _control.moveProbeY (increasing ? -1. : 1.);
        _control.resetYTipStrain();
        
        for (double x = saveX; x < saveX + maxDeltaX; x += SWEEP_TO_CONTACT_INCREMENT) {
            // Prepare probe position for before we move
            _control.moveProbeY (increasing ? -1. : 1.);

            _control.moveToXY (x, saveY);
            if (moveProbeYToContact (increasing))
                // contacted the block
                return (true);
            CancelThread.check();
        }
        
        return (false);
    }
    
    /** Move the stage in the X direction until the probe X tip strain gauge detects
     * a contact force,
     * then back off a little until the force drops near zero.  If increasing
     * is true then the position increases (as defined by moveToXY), otherwise decreases.
     * If the stage moves
     * the limit distance without contact, then return false.  To do a similar
     * operation by moving the probe instead of the stage, use moveProbeXToContact.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param increasing    true to increase the X position until contact,
     *                      false to decrease
     * @param limit     the maximum distance in millimeters to move while
     *                  checking for contact. This should be a positive number
     *          (even if the stage will move in the negative direction).  If
     *          it is a negative number, this takes its absolute value.
     * @return  true for success, false if couldn't find contact before the
     *          limit distance was reached
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     * @see     #moveProbeXToContact
     */
    public boolean moveStageXToContact (boolean increasing, double limit)
    throws CancelException {
        // make sure limit is a positive number
        limit = Math.abs (limit);
        double sign = (increasing ? 1. : -1);
        boolean found = false;
        double startingX = _control.stageX();
        for (double x = startingX;
             increasing ? x < startingX + limit : x > startingX - limit;
             x += sign * MOVE_STAGE_CONTACT_INCREMENT) {
            _control.moveToXY (x, _control.stageY());
            CancelThread.check();

            if (increasing ? _control.xTipStrain() < -3. * _control.strainErrorXTip() :
                             _control.xTipStrain() >  3. * _control.strainErrorXTip()) {
                found = true;
                break;
            }
        }
        if (!found) {
            // couldn't find initial contact
            return false;
        }

        // Now back off on until the strain drops back near zero
        double saveX = _control.stageX();
        for (double x = saveX;
             increasing ? x > saveX - limit : x < saveX + limit;
             x += -sign * MOVE_STAGE_CONTACT_INCREMENT) {
            _control.moveToXY (x, _control.stageY());
            CancelThread.check();

            if (increasing ? _control.xTipStrain() > -_control.strainErrorXTip() :
                             _control.xTipStrain() <  _control.strainErrorXTip()) {
                // Successfully got the contact to release
                return true;
            }
        }

        // Could not get reading to fall back to zero.  (This really shouldn't happen)
        return false;
    }

    /** Move the stage in the Y direction until the probe Y tip strain gauge detects
     * a contact force,
     * then back off a little until the force drops near zero.  If increasing
     * is true then the position increases (as defined by moveToXY), otherwise decreases.
     * If the stage moves
     * the limit distance without contact, then return false.  To do a similar
     * operation by moving the probe instead of the stage, use moveProbeYToContact.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param increasing    true to increase the X position until contact,
     *                      false to decrease
     * @param limit     the maximum distance in millimeters to move while
     *                  checking for contact.  This should be a positive number
     *          (even if the stage will move in the negative direction).  If
     *          it is a negative number, this takes its absolute value.
     * @return  true for success, false if couldn't find contact before the
     *          limit distance was reached
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     * @see     #moveProbeYToContact
     */
    public boolean moveStageYToContact (boolean increasing, double limit)
          throws CancelException {
        try {
            // make sure limit is a positive number
            limit = Math.abs (limit);
            double sign = (increasing ? 1. : -1);
            boolean found = false;
            double startingY = _control.stageY();
            for (double y = startingY;
                 increasing ? y < startingY + limit : y > startingY - limit;
                 y += sign * MOVE_STAGE_CONTACT_INCREMENT) {
                _control.moveToXY (_control.stageX(), y);
                CancelThread.check();

                // Because the stage shakes the probes, wait for it to settle
                Thread.sleep (ARM_SETTLING_MILLIS);
                CancelThread.check();

                if (increasing ? _control.yTipStrain() < -3. * _control.strainErrorYTip() :
                                 _control.yTipStrain() >  3. * _control.strainErrorYTip()) {
                    found = true;
                    break;
                }
            }
            if (!found) {
                // couldn't find initial contact
                return false;
            }

            // Now back off on until the strain drops back near zero
            double saveY = _control.stageY();
            for (double y = saveY;
                 increasing ? y > saveY - limit : y < saveY + limit;
                 y += -sign * MOVE_STAGE_CONTACT_INCREMENT) {
                _control.moveToXY (_control.stageX(), y);
                CancelThread.check();

                // Probe is in contact so the arms should not vibrate due to
                //   the stage moving so no need to wait for it to settle.
                if (increasing ? _control.yTipStrain() > -_control.strainErrorYTip() :
                                 _control.yTipStrain() <  _control.strainErrorYTip()) {
                    // Successfully got the contact to release
                    return true;
                }
            }

            // Could not get reading to fall back to zero.  (This really shouldn't happen)
            return false;
        } catch (InterruptedException e) {
            throw new CancelException ("interrupted exception");
        }
    }

    /** Move the stage in the negative Z direction, making tiny sweeps with probe X,
     * until the probe X tip strain gauge detects a contact force when it is moved,
     * Then increase Z until no force is detected.
     * This can be used to find the Z coordinate of the surface.  This assumes that
     * initially probeX is not in contact with anything, especially probe Y.
     * If the stage moves the limit distance without contact, then return false.  
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param limit     the maximum distance in millimeters to move while
     *                  checking for contact. This should be a positive number
     *          (even if the stage moves in the negative Z direction).  If
     *          it is a negative number, this takes its absolute value.
     * @return  true for found the surface, false if couldn't find contact before the
     *          limit distance was reached, or if couldn't release the contact force
     *          after it was encountered.
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     * @see     #moveProbeXToContact
     */
    public boolean seekZSurface (double limit)
    throws CancelException {
        // make sure limit is a positive number
        limit = Math.abs (limit);
        double startingZ = _control.stageZ();
        // Move the arm enough outward to detect the change above the error level
        double armDeflection = Math.abs(2. * _control.strainErrorXArm() * _control.strainToDeflectionXArm());
        double minDeflection = _control.xArmDeflection();
        double maxDeflection = minDeflection + armDeflection;
        
        boolean increasing = true;
        boolean found = false;
        for (double z = startingZ; z > startingZ - limit; z -= 5 * MOVE_STAGE_CONTACT_INCREMENT){
            _control.moveToZ (z);

            // moveProbeXToContactBounded returns false if there is already
            // contact but in the opposite direction, so check first if there
            // is contact
            double strain = _control.xTipStrain();
            if (strain <= -3 * _control.strainErrorXTip() || 
                strain >= 3 * _control.strainErrorXTip()) {
                found = true;
                break;
            }
            
            if (moveProbeXToContactBounded (increasing, minDeflection, maxDeflection)) {
                found = true;
                break;
            }

            // Sweep the other direction next time
            increasing = !increasing;
        }
        
        if (!found)
            // couldn't find contact
            return false;

        // The arm should be in contact in the direction indicated by increasing.
        // Apply a force, and then increase Z until no force is felt.
        seekXArmDeflection (_control.xArmDeflection() + (increasing ? armDeflection : -armDeflection));

        double saveZ = _control.stageZ();
        for (double z = saveZ; z < saveZ + limit; z += MOVE_STAGE_CONTACT_INCREMENT){
            _control.moveToZ (z);

            double strain = _control.xTipStrain();
            if (strain >= -_control.strainErrorXTip() && strain <= _control.strainErrorXTip())
                // Achived zero strain, within the error limits
                return true;
        }

        // couldn't release
        return false;
    }

    /** This is a simplified version of seekXTipForceBounded
     * which allows the probe to seek for the entire probe range.
     *
     * @see         #seekXTipForceBounded
     */
    public boolean seekXTipForce (double force) throws CancelException {
        // Use rediculously large values for minDeflection and maxDeflection
        // so that this will limit to the min and max positions of -1 or 1.
        return seekXTipForceBounded (force, -1000., 1000.);
    }
    
    /** Move probe X, bounded to the given range, until its tip force is the given value. This seeks until the
     * desired force, or until the probe deflection reaches minDeflection or
     * maxDeflection, or until the probe position reaches the absolute limits
     * of position -1 or 1.  This uses
     * feedback from the tip (but not arm) strain gauge.  The direction in which
     * the probe will move is derived from comparing the present force on the
     * tip with the desired force.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param force    the force to seek to in milli Newtons
     * @param minDeflection minimum probe X deflection (in millimeters) allowed 
     *                  when seeking the desired force. This also stops if 
     *                  control.probeXPosition() reaches the absolute limit of -1.
     *                  If minDeflection >= maxDeflection, this simply returns false.
     * @param maxDeflection maximum probe X deflection (in millimeters) allowed 
     *                  when seeking the desired force. This also stops if 
     *                  control.probeXPosition() reaches the absolute limit of +1.
     *                  If maxDeflection <= minDeflection, this simply returns false.
     * @return  true if achieved the requested force, false if not in which
     *          case the probe has either reached minDeflection or maxDeflection,
     *          or the absolute position limits of -1 or 1.
     * @exception CancelException   thrown by CancelThread.check()
     *              or if got an InterruptedException.
     * @see OrthoTweezersControl#probeXPosition
     * @see         CancelThread
     */
    public boolean seekXTipForceBounded (double force, double minDeflection, double maxDeflection) 
           throws CancelException {
        if (minDeflection >= maxDeflection)
            // invalid range
            return false;
        
        try {
            double forceError = _control.strainToForceXTip() * _control.strainErrorXTip();
            if (force == 0.)
                // When aiming for a force of 0., we are not pushing against
                //   something that will allow us to adjust the force, so
                //   we must expand the allowable error region.
                forceError *= 3.;
            
            // Do this whole algorithm twice, the first time to get in the
            //   general neighborhood, then wait a little for the probe dynamics
            //   to settle down and then seek again to get closer.
            for (int i = 0; i < 2; ++i) {
                // arm position must increase to make the tip force go more negative
                boolean increasing = (force < _control.xTipForce());
                double sign = (increasing ? 1. : -1);
                
                boolean found = false;
                double position = _control.probeXPosition();
                while (true) {
                    // When testing the tip force, allow success if within force error
                    if (increasing ? _control.xTipForce() < (force + forceError)
                                   : _control.xTipForce() > (force - forceError)) {
                       found = true;
                       break;
                    }
                
                    // The first time through, step by more than SEEK_POSITION_INCREMENT to go faster
                    position += sign * (i == 0 ? 3 : 1) * SEEK_POSITION_INCREMENT;
                    if (increasing ? (position >= 1. || _control.xArmDeflection() >= maxDeflection) 
                        : (position <= -1 || _control.xArmDeflection() <= minDeflection)) {
                        // Reached the limit of motion
                        break;
                    }

                    _control.moveProbeX (position);
                    CancelThread.check();
                }
                
                if (!found)
                   return false;

                if (i == 0)
                   Thread.sleep (ARM_SETTLING_MILLIS);
            }
        } catch (InterruptedException e) {
            throw new CancelException ("interrupted exception");
        }
            
        return true;
    }

    /** This is a simplified version of seekYTipForceBounded
     * which allows the probe to seek for the entire probe range.
     *
     * @see         #seekYTipForceBounded
     */
    public boolean seekYTipForce (double force) throws CancelException {
        // Use rediculously large values for minDeflection and maxDeflection
        // so that this will limit to the min and max positions of -1 or 1.
        return seekYTipForceBounded (force, -1000., 1000.);
    }
    
    /** Move probe Y, bounded to the given range, until its tip force is the given value. This seeks until the
     * desired force, or until the probe deflection reaches minDeflection or
     * maxDeflection, or until the probe position reaches the absolute limits
     * of position -1 or 1.  This uses
     * feedback from the tip (but not arm) strain gauge.  The direction in which
     * the probe will move is derived from comparing the present force on the
     * tip with the desired force.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param force    the force to seek to in milli Newtons
     * @param minDeflection minimum probe Y deflection (in millimeters) allowed 
     *                  when seeking the desired force. This also stops if 
     *                  control.probeYPosition() reaches the absolute limit of -1.
     *                  If minDeflection >= maxDeflection, this simply returns false.
     * @param maxDeflection maximum probe Y deflection (in millimeters) allowed 
     *                  when seeking the desired force. This also stops if 
     *                  control.probeYPosition() reaches the absolute limit of +1.
     *                  If maxDeflection <= minDeflection, this simply returns false.
     * @return  true if achieved the requested force, false if not in which
     *          case the probe has either reached minDeflection or maxDeflection,
     *          or the absolute position limits of -1 or 1.
     * @exception CancelException   thrown by CancelThread.check()
     *              or if got an InterruptedException.
     * @see OrthoTweezersControl#probeYPosition
     * @see         CancelThread
     */
    public boolean seekYTipForceBounded (double force, double minDeflection, double maxDeflection) 
           throws CancelException {
        if (minDeflection >= maxDeflection)
            // invalid range
            return false;
        
        try {
            double forceError = _control.strainToForceYTip() * _control.strainErrorYTip();
            if (force == 0.)
                // When aiming for a force of 0., we are not pushing against
                //   something that will allow us to adjust the force, so
                //   we must expand the allowable error region.
                forceError *= 3.;
            
            // Do this whole algorithm twice, the first time to get in the
            //   general neighborhood, then wait a little for the probe dynamics
            //   to settle down and then seek again to get closer.
            for (int i = 0; i < 2; ++i) {
                // arm position must increase to make the tip force go more negative
                boolean increasing = (force < _control.yTipForce());
                double sign = (increasing ? 1. : -1);
                
                boolean found = false;
                double position = _control.probeYPosition();
                while (true) {
                    // When testing the tip force, allow success if within force error
                    if (increasing ? _control.yTipForce() < (force + forceError)
                                   : _control.yTipForce() > (force - forceError)) {
                       found = true;
                       break;
                    }
                
                    // The first time through, step by more than SEEK_POSITION_INCREMENT to go faster
                    position += sign * (i == 0 ? 3 : 1) * SEEK_POSITION_INCREMENT;
                    if (increasing ? (position >= 1. || _control.yArmDeflection() >= maxDeflection) 
                                   : (position <= -1 || _control.yArmDeflection() <= minDeflection))
                        // Reached the limit of motion
                        break;

                    _control.moveProbeY (position);
                    CancelThread.check();
                }
                
                if (!found)
                   return false;

                if (i == 0)
                   Thread.sleep (ARM_SETTLING_MILLIS);
            }
        } catch (InterruptedException e) {
            throw new CancelException ("interrupted exception");
        }
            
        return true;
    }

    /** Grasp a rectangular object.  The object to grasp must be in the path
     * of both probe X and probe Y as they would close inward.  To achieve
     * this, you may want to use sweepProbeXToContact and sweepProbeYToContact.
     * This assumes that the path is free for both
     * probe arms to defelect fully outward. After deflecting both arms fully
     * outward, this calls control().resetXTipStrain() and control().resetYTipStrain()
     * to recalibrate the zero offset of the probes so that they are sensitive.
     * This sweeps at the present
     * Z height in the decreasing X and Y directions along the edges of the
     * rectangular object until finding the lower and left corners of the object,
     * which is assumed to be at angle (rather than aligned to the X-Y axes).
     * Once the corners are found, this computes the grip points which are
     * a distance of cornerOffset along the edges from the corner. Then this
     * grips with the given gripForce.
     * 
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param cornerOffset    the distance along the edge from the corners
     *                      to graps, in millimeters.  If this is less than
     *                      0.01, this sets it to 0.01.
     * @param gripForce     the force with which to grip, in milli Newtons
     * @return  null for success, otherwise an error String describing the problem.
     * @exception CancelException   thrown by CancelThread.check()
     *              or if got an InterruptedException.
     * @see         CancelThread
     * @see #sweepProbeXToContact
     * @see #sweepProbeYToContact
     */
    public String grasp (double cornerOffset, double gripForce) throws CancelException {
        if (cornerOffset < 0.01)
            cornerOffset = 0.01;
        if (gripForce < 0.)
            gripForce = 0.;
        
        // Make sure probe Y is out of the way
        _control.moveProbeY (1.);
        // Reset strain so that reading is sensitive.
        _control.resetYTipStrain();
        double saveX = _control.stageX();
        double saveY = _control.stageY();

        if (!moveProbeXToContact (false))
            return ("ERROR: Object is not already in path of probe X");
        Point2d lowerCorner = new Point2d ();
        lowerCorner.x = saveX + _control.xArmDeflection();
        lowerCorner.y = saveY;
        Point2d initialLowerCorner = new Point2d(lowerCorner);
        
        // back off until we miss the block
        boolean found = false;
        for (double y = saveY; y >= saveY - .5; y -= .005) {
            // Move the probe a little to sweep again
            moveProbeDeltaX (.1);
            CancelThread.check();
            _control.moveToXY (saveX, y);
            // Use bounded motion of the sweep to the left because we shouldn't
            //   have to sweep all the way.  A value of 1000 on the right will be clipped 
            //   to the maximum range.
            // Present deflection is lowerCorner.x - saveX
            if (moveProbeXToContactBounded (false, lowerCorner.x - saveX - .1, 1000) == false) {
                // missed the block
                found = true;
                break;
            }
            // Contacted the block again, so revise lowerCorner
            lowerCorner.x = _control.stageX() + _control.xArmDeflection();
            lowerCorner.y = y;
        }
        if (!found) {
            // can't contact the block.  Perhaps Z is too high, or we should
            //   move over in X or farther in Y and try again.
            return ("ERROR: Can't find block lower corner");
        }
        
        // Now do the same operation with probe Y to find the left corner.
        // First move a little in the Y direction so probeX certainly won't hit.
        saveY = lowerCorner.y - .01;
        _control.moveToXY (saveX, saveY);
        _control.moveProbeX (1.);
        // Reset strain so that reading is sensitive.
        _control.resetXTipStrain();
        
        if (!moveProbeYToContact (false))
            return ("ERROR: Object is not already in path of probe Y");
        CancelThread.check();
        Point2d leftCorner = new Point2d();;
        leftCorner.x = saveX;
        leftCorner.y = _control.stageY() + _control.yArmDeflection();
        Point2d initialLeftCorner = new Point2d(leftCorner);
        
        // back off until we miss the block
        found = false;
        for (double x = saveX; x >= saveX - .5; x -= .005) {
            // Move the probe a little to sweep again
            moveProbeDeltaY (.1);
            CancelThread.check();

            _control.moveToXY (x, saveY);
            // Use bounded motion of the sweep to the bottom because we shouldn't
            //   have to sweep all the way.  A value of 1000 on the top will be clipped 
            //   to the maximum range.
            // Present deflection is lowerCorner.y - saveY
            if (moveProbeYToContactBounded (false, leftCorner.y - saveY - .1, 1000) == false) {
                // missed the block
                found = true;
                break;
            }
            // Contacted the block again, so revise lowerCorner
            leftCorner.x = x;
            leftCorner.y = _control.stageY() + _control.yArmDeflection();
        }
        if (!found) {
            // can't contact the block.  Perhaps Z is too high, or we should
            //   move over in X or farther in Y and try again.
            return ("ERROR: Can't find block left corner");
        }
        
        // Get a vector lowerRightSide which should point in the direction along the lower
        //   right side up from lowerCorner, which is the same as pointing along the upper
        //   left side from leftCorner.  Do this by setting lowerLeftSide first.
        Vector2d lowerLeftSide = new Vector2d (leftCorner);
        lowerLeftSide.sub (lowerCorner);
        lowerLeftSide.normalize();
        // lowerRightSide is lowerLeftSide rotated by -90 degrees
        Vector2d lowerRightSide = new Vector2d (lowerLeftSide.y, -lowerLeftSide.x);
        
        // Now get the points of contact which are cornerOffset from the corners
        Point2d probeXContact = new Point2d();
        Point2d probeYContact = new Point2d();
        boolean graspAtRightAngle = false;
        double lowerRightSideAngle = Math.atan2(lowerLeftSide.y, lowerLeftSide.x);
        if (lowerRightSideAngle < 0. || lowerRightSideAngle > Math.PI/2.) {
            // Special case.  Something is strange about the point contacts such
            // that the do not point in the expected range of angles.  
            // Point the contacts off at 45 degrees and try to use the grasp
            // at right angles algorithm.
            lowerRightSide.set (1., 1.);
            lowerRightSide.normalize();
            graspAtRightAngle = true;
        }
        probeXContact.scaleAdd (cornerOffset, lowerRightSide, lowerCorner);
        probeYContact.scaleAdd (cornerOffset, lowerRightSide, leftCorner);
        
        // first, move to clear the block and position probes out of the way
        _control.moveToXY (leftCorner.x - .01, lowerCorner.y - .01);
        _control.moveProbeX (1.);
        _control.moveProbeY (1.);
        CancelThread.check();
        
        // Position to contact and move the probes in
        _control.moveToXY (probeYContact.x, probeXContact.y);
        if (!moveProbeXToContact (false))
            return "Cannot contact object at probe X contact";
        if (!moveProbeYToContact (false))
            return "Cannot contact object at probe Y contact";
        CancelThread.check();

        // Get actual points of contact and compare with desired
        Point2d actualProbeXContact = new Point2d
          (_control.stageX() + _control.xArmDeflection(), _control.stageY());
        Point2d actualProbeYContact = new Point2d
          (_control.stageX(), _control.stageY() + _control.yArmDeflection());
        
        // Use seekZAxisAngle at the present angle as a way of getting to the desired force.
        seekZAxisAngle (zAxisAngle(), gripForce);

        // finished
        return null;
    }    

    /** Return the angle in radians of the grip axis as it would rotate around
     * the Z axis.  The grip axis is a line between the tips of the two probes,
     * where the position of the tip is found by adding the probe arm and tip
     * deflections.
     * An angle of zero is when probe X is at zero deflection and
     * probe Y is at some positive deflection.  An angle of PI/2 is when
     * probe X is at some positive deflection and probe Y is at zero deflection.
     * In this way, a movement from angle zero to angle PI/2 corresponds to
     * a part rotating in the positive direction around the Z axis.
     */
    public double zAxisAngle() {
        double probeXDeflection = _control.xArmDeflection() + _control.xTipDeflection();
        double probeYDeflection = _control.yArmDeflection() + _control.yTipDeflection();
        return Math.atan2 (probeXDeflection, probeYDeflection);
    }
    
    /** Try to rotate a block about the Z axis to the given angle.
     * This assumes that both probes are in contact with a block which can
     * be rotated freely.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @param angle     the angle to rotate to in radians.  This is measured by the relative
     *          positions of the probe tips.  When probe X is at deflection zero
     *          and probe Y has a positive deflection, this angle is zero.
     *          The angle increases to PI/2 where probe Y has zero deflection and probe
     *          X has a positive deflection.  If angle is less than zero or greater than
     *          PI/2, it is clipped to this range.
     * @param gripForce     the gripping force in milliNewtons to maintain while rotating
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     */
    public void seekZAxisAngle (double angle, double gripForce) throws CancelException {
        if (angle < 0.)
            angle = 0.;
        if (angle > Math.PI / 2.)
            angle = Math.PI / 2.;
        
        double positionIncrement = 5. * SEEK_POSITION_INCREMENT;
        
        // DEBUG: should have a better failure check to avoid infinite loop.
        for (int i = 0; i < 100; ++i) {
            boolean increasing = (angle > zAxisAngle());
            
            if (_control.xTipForce() < gripForce && _control.yTipForce() < gripForce) {
                // Both probes are under the gripping force, so increase the force (decrease
                //   the deflection) on the probe which would cause rotation in the desired
                //   direction.
                if (increasing) {
                    //seekYArmDeflection (yArmDeflection - deflectionIncrementY);
                    _control.moveProbeY (_control.probeYPosition() - positionIncrement);
                }
                else {
                    //seekXArmDeflection (xArmDeflection - deflectionIncrementX);
                    _control.moveProbeX (_control.probeXPosition() - positionIncrement);
                }
            }
            else {
                // One of the probes is over the gripping force, so decrease the force (increase
                //   the deflection) on the probe which would cause rotation in the desired
                //   direction.
                if (increasing) {
                    //seekXArmDeflection (xArmDeflection + deflectionIncrementX);
                    _control.moveProbeX (_control.probeXPosition() + positionIncrement);
                }
                else {
                    //seekYArmDeflection (yArmDeflection + deflectionIncrementY);
                    _control.moveProbeY (_control.probeYPosition() + positionIncrement);
                }
            }
        }
    }
    
    /** Ungrasp an object by releasing both probes to zero force, moving down in
     * Z slightly, opening the probes slightly, and restoring the Z position.
     * If the zAxisAngle() of the part is near zero or PI/2, then on of the probes
     * must be released by moving the stage (and the object with it) away from the
     * probe.
     * This calls CancelThread.check() to maybe throw a CancelException.
     *
     * @exception CancelException   thrown by CancelThread.check()
     * @see         CancelThread
     * @see #zAxisAngle
     */
    public void ungrasp() throws CancelException {
        double angle = zAxisAngle();
        // Save deflections because they could be modified by seekZAxisAngle.
        double xDeflection = _control.xArmDeflection() + _control.xTipDeflection();
        double yDeflection = _control.yArmDeflection() + _control.yTipDeflection();

        // Use seekZAxisAngle at the present angle as a way of decreasing force.
        seekZAxisAngle (angle, 0.);
        // This should loosen the grip
        moveDeltaZ(-.01);
        
        if (angle < Math.PI/10.) {
            // Special case for when object is almost at right angles.
            // Release the Y probe by moving it, but release the X probe by
            //   moving the stage.
            seekYArmDeflection(yDeflection + .08);
            moveDeltaXY(0., -.04);
        }
        else if (angle > Math.PI*4./10.) { 
            // Special case for when object is almost at right angles.
            // Release the X probe by moving it, but release the Y probe by
            //   moving the stage.
            seekXArmDeflection(xDeflection + .08);
            moveDeltaXY(-.04, 0.);
        }
        else {
            seekXArmDeflection(xDeflection + .08);
            seekYArmDeflection(yDeflection + .08);
        }
        
        moveDeltaZ(.01);
    }
    
    /** Turn on the heater specified by heaterNumber.
     *
     * @param heaterNumber - the number of the heater to turn on. At present
     *      there is only one heater on the system so this number must be 1.
     *      In the future there may be heater number 2, 3, etc.  If heaterNumber
     *      is not 1, this does nothing.
     */
    public void heaterOn (int heaterNumber) {
        if (heaterNumber == 1)
            _control.relay1On();
    }
    
    /** Turn off the heater specified by heaterNumber.
     *
     * @param heaterNumber - the number of the heater to turn off. At present
     *      there is only one heater on the system so this number must be 1.
     *      In the future there may be heater number 2, 3, etc.  If heaterNumber
     *      is not 1, this does nothing.
     */
    public void heaterOff (int heaterNumber) {
        if (heaterNumber == 1)
            _control.relay1Off();
    }
    
    /** Return the state of the heater specified by heaterNumber.
     *
     * @param heaterNumber - the number of the heater to turn off. At present
     *      there is only one heater on the system so this number must be 1.
     *      In the future there may be heater number 2, 3, etc.  If heaterNumber
     *      is not 1, this does nothing and returns false.
     * @return true if the heater is on, otherwise false.
     */
    public boolean heaterState (int heaterNumber) {
        if (heaterNumber == 1)
            return _control.relay1State();
        
        return false;
    }
    
    // For use in sweepProbeXToContact, etc.
    public static final double SWEEP_TO_CONTACT_INCREMENT = .050;
    final static long ARM_SETTLING_MILLIS = 50;
    // For use in moveStageXToContact, etc.  This determines the accuracy
    //   of the contact measurement
    final static double MOVE_STAGE_CONTACT_INCREMENT = .001;
    // For use in seekProbeXDeflection, etc.
    final static double SEEK_POSITION_INCREMENT = .001;

    private OrthoTweezersControl _control;
}
    
