/*
Copyright (c) 2000,2001. The Regents of the University of California
(Regents). All Rights Reserved. 

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research, and not-for-profit purposes,
without fee and without a signed licensing agreement, is hereby
granted, provided that the above copyright notice, this paragraph and
the following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC
Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620,
(510) 643-7201, for commercial licensing opportunities.
  
IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. 
  
REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF
ANY, PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO
OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.
*/
				 
package orthotweezers;

import java.io.*;
import java.util.*;

/** Preferences keeps a list of key/value pairs where each is a String.  
 * It has methods to read and write the preferences file.
 *
 * @author  unknown
 */
public class Preferences extends Properties {
    /** Creates a new Preferences.
     * To load values from a preferences file, call loadFromFile.
     *
     * @see #loadFromFile
     */
    public Preferences() {
        super();
    }
    
    /** Open the given file and load the preferences.
     * To get the value call getProperty(), or to interpret
     * the value, call a method like getDouble().  This uses Properties.load()
     * to load the file.
     *
     * @return  true for success, false for error opening file.  Errors
     *          are printed to System.out
     * @see #getProperty
     * @see #getDouble
     */
    public boolean loadFromFile (File file) {
        try {
            BufferedInputStream input = 
              new BufferedInputStream (new FileInputStream (file));

            load (input);
            input.close();
        } catch (Exception e) {
            System.out.println ("Preferences reading: " + e);
            return false;
        }

        return true;
    }

    /** Save the preferences to the given file.  The preferences are all
     * the key/value strings read in by loadFromFile() or added by put().
     * The file is written alphabetically by key.  This uses Properties.store()
     * to save the file.
     *
     * @return  true for success, false for error opening file.  Errors are
     * printed to System.out.
     */
    public boolean saveToFile (File file) {
        try {
            BufferedOutputStream output = new BufferedOutputStream (new FileOutputStream (file));
            
            store (output, "OrthoTweezers preferences");
            output.close();
        } catch (Exception e) {
            System.out.println ("Preferences writing: " + e);
            return false;
        }

        return true;
    }
    
    /** Return the value of the preferences with the given key, parsing
     * the value as a Double.  If the preference with the key is absent, or the
     * value cannot be interpreted as a double, return null.  If the value cannot
     * be parsed, this prints a warning to System.out.
     * For the key, you can use provided String constants such as STAGE_X.
     * 
     * @param key   the key for the value to get.  The key is case sensitive.
     * @return  the value of the given key as a Double object or null if
     *          the key is not found or cannot parse the value as a double
     * @see #maybeGetDouble
     */
    public Double getDouble (String key) {
        try {
            String value = getProperty (key);
            if (value == null)
                return (null);
            
            return (new Double (value));
        } catch (NumberFormatException e) {
            // value string is not a number
            System.out.println 
              ("WARNING: Preferences: value of " + key + " is not a number.");
            
            return (null);
        }
    }
    
    /** Return the value of the preferences with the given key, parsing
     * the value as a double.  If the preference with the key is absent, or the
     * value cannot be interpreted as a double, return previous.  If the value cannot
     * be parsed, this prints a warning to System.out.
     * <p>The idea is that you can write:
     * <p>param = preferences.maybeGetDouble (key, param); 
     * <p>and param will only be changed if the key is in the preferences.
     * For the key, you can use provided String constants such as STAGE_X.
     * 
     * @param key   the key for the value to get.  The key is case sensitive.
     * @param previous  the value to return in case the key is absent or the
     *                  value cannot be interpreted as a double
     * @return  the value of the given key as a Double object or previous if
     *          the key is not found or cannot parse the value as a double
     */
    public double maybeGetDouble (String key, double previous) {
        Double value = getDouble (key);
        if (value == null)
            return previous;
        else
            return value.doubleValue();
    }

    /** Puts the preference with the given key and value in the preferences.
     * If a preference with the key already exists, this replaces it.
     * Note, for typical preference keys, you should define and
     * use String constants so you don't have to type the String separately each time.
     * For the key, you can use provided String constants such as STAGE_X.
     *
     * @param key   the key for the value to put.  The key is case sensitive.
     */
    public void putDouble (String key, double value) {
        setProperty (key, Double.toString (value));
    }

    // String constants for the preference keys.
    public static final String STAGE_X = "stagex";
    public static final String STAGE_Y = "stagey";
    public static final String STAGE_Z = "stagez";
    
    public static final String STRAIN_ERROR_X_ARM = "strainerrorxarm";
    public static final String STRAIN_ERROR_Y_ARM = "strainerroryarm";
    public static final String STRAIN_ERROR_X_TIP = "strainerrorxtip";
    public static final String STRAIN_ERROR_Y_TIP = "strainerrorytip";

    public static final String STRAIN_OFFSET_X_ARM = "strainoffsetxarm";
    public static final String STRAIN_OFFSET_Y_ARM = "strainoffsetyarm";
    public static final String STRAIN_OFFSET_X_TIP = "strainoffsetxtip";
    public static final String STRAIN_OFFSET_Y_TIP = "strainoffsetytip";

    public static final String STRAIN_TO_DEFLECTION_X_ARM = "straintodeflectionxarm";
    public static final String STRAIN_TO_DEFLECTION_Y_ARM = "straintodeflectionyarm";
    public static final String STRAIN_TO_DEFLECTION_X_TIP = "straintodeflectionxtip";
    public static final String STRAIN_TO_DEFLECTION_Y_TIP = "straintodeflectionytip";
    public static final String STRAIN_TO_FORCE_X_TIP = "straintoforcextip";
    public static final String STRAIN_TO_FORCE_Y_TIP = "straintoforceytip";
}